import copy
import errno
import os
import socket
import struct
import threading
from itertools import count

from pyroute2.config import AF_NETLINK
from pyroute2.netlink import (
    NLM_F_DUMP,
    NLM_F_MULTI,
    NLMSG_DONE,
    nlmsg,
    nlmsgerr,
)
from pyroute2.netlink.core import Stats
from pyroute2.netlink.rtnl import (
    RTM_DELADDR,
    RTM_DELROUTE,
    RTM_GETADDR,
    RTM_GETLINK,
    RTM_GETROUTE,
    RTM_NEWADDR,
    RTM_NEWLINK,
    RTM_NEWROUTE,
)
from pyroute2.netlink.rtnl.ifaddrmsg import ifaddrmsg
from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg
from pyroute2.netlink.rtnl.marshal import MarshalRtnl
from pyroute2.netlink.rtnl.rtmsg import rtmsg

interface_counter = count(3)
MAGIC_CLOSE = 0x42


class MockLink:
    def __init__(
        self,
        index=0,
        ifname='',
        address='00:00:00:00:00:00',
        broadcast='ff:ff:ff:ff:ff:ff',
        perm_address=None,
        flags=1,
        rx_bytes=0,
        tx_bytes=0,
        rx_packets=0,
        tx_packets=0,
        mtu=0,
        qdisc='noqueue',
        kind=None,
        link=None,
        vlan_id=None,
        master=0,
        br_max_age=0,
        br_forward_delay=0,
        alt_ifname_list=None,
    ):
        self.index = index if index > 0 else next(interface_counter)
        self.ifname = ifname
        self.flags = flags
        self.address = address
        self.broadcast = broadcast
        self.perm_address = perm_address
        self.rx_bytes = rx_bytes
        self.tx_bytes = tx_bytes
        self.rx_packets = rx_packets
        self.tx_packets = tx_packets
        self.mtu = mtu
        self.qdisc = qdisc
        self.kind = kind
        self.link = link
        self.vlan_id = vlan_id
        self.master = master
        self.br_max_age = br_max_age
        self.br_forward_delay = br_forward_delay
        self.alt_ifname_list = alt_ifname_list or []

    def update_from_msg(self, msg):
        [
            setattr(self, x, msg.get(x))
            for x in ['address', 'broadcast', 'mtu', 'ifname', 'master']
            if msg.get(x) is not None
        ]
        #
        self.kind = msg.get(('linkinfo', 'kind'))
        if msg.get('change') != 0:
            self.flags = msg.get('flags')
        # vlan
        if self.kind == 'vlan':
            self.vlan_id = msg.get(('linkinfo', 'data', 'vlan_id'))
            self.link = msg.get('link')
        elif self.kind == 'bridge':
            self.br_max_age = msg.get(('linkinfo', 'data', 'br_max_age'))
            self.br_forward_delay = msg.get(
                ('linkinfo', 'data', 'br_forward_delay')
            )

    @classmethod
    def load_from_msg(cls, msg):
        ret = cls()
        ret.update_from_msg(msg)
        return ret

    def export(self):
        ret = {
            'attrs': [
                ['IFLA_IFNAME', self.ifname],
                ['IFLA_TXQLEN', 1000],
                ['IFLA_OPERSTATE', 'UNKNOWN'],
                ['IFLA_LINKMODE', 0],
                ['IFLA_MTU', self.mtu],
                ['IFLA_GROUP', 0],
                ['IFLA_PROMISCUITY', 0],
                ['IFLA_NUM_TX_QUEUES', 1],
                ['IFLA_GSO_MAX_SEGS', 65535],
                ['IFLA_GSO_MAX_SIZE', 65536],
                ['IFLA_GRO_MAX_SIZE', 65536],
                ['IFLA_NUM_RX_QUEUES', 1],
                ['IFLA_CARRIER', 1],
                ['IFLA_QDISC', self.qdisc],
                ['IFLA_CARRIER_CHANGES', 0],
                ['IFLA_CARRIER_UP_COUNT', 0],
                ['IFLA_CARRIER_DOWN_COUNT', 0],
                ['IFLA_PROTO_DOWN', 0],
                [
                    'IFLA_MAP',
                    {
                        'base_addr': 0,
                        'dma': 0,
                        'irq': 0,
                        'mem_end': 0,
                        'mem_start': 0,
                        'port': 0,
                    },
                ],
                ['IFLA_ADDRESS', self.address],
                ['IFLA_BROADCAST', self.broadcast],
                [
                    'IFLA_STATS64',
                    {
                        'collisions': 0,
                        'multicast': 0,
                        'rx_bytes': self.rx_bytes,
                        'rx_compressed': 0,
                        'rx_crc_errors': 0,
                        'rx_dropped': 0,
                        'rx_errors': 0,
                        'rx_fifo_errors': 0,
                        'rx_frame_errors': 0,
                        'rx_length_errors': 0,
                        'rx_missed_errors': 0,
                        'rx_over_errors': 0,
                        'rx_packets': self.rx_packets,
                        'tx_aborted_errors': 0,
                        'tx_bytes': self.tx_bytes,
                        'tx_carrier_errors': 0,
                        'tx_compressed': 0,
                        'tx_dropped': 0,
                        'tx_errors': 0,
                        'tx_fifo_errors': 0,
                        'tx_heartbeat_errors': 0,
                        'tx_packets': self.tx_packets,
                        'tx_window_errors': 0,
                    },
                ],
                [
                    'IFLA_STATS',
                    {
                        'collisions': 0,
                        'multicast': 0,
                        'rx_bytes': self.rx_bytes,
                        'rx_compressed': 0,
                        'rx_crc_errors': 0,
                        'rx_dropped': 0,
                        'rx_errors': 0,
                        'rx_fifo_errors': 0,
                        'rx_frame_errors': 0,
                        'rx_length_errors': 0,
                        'rx_missed_errors': 0,
                        'rx_over_errors': 0,
                        'rx_packets': self.rx_packets,
                        'tx_aborted_errors': 0,
                        'tx_bytes': self.tx_bytes,
                        'tx_carrier_errors': 0,
                        'tx_compressed': 0,
                        'tx_dropped': 0,
                        'tx_errors': 0,
                        'tx_fifo_errors': 0,
                        'tx_heartbeat_errors': 0,
                        'tx_packets': self.tx_packets,
                        'tx_window_errors': 0,
                    },
                ],
                ['IFLA_XDP', {'attrs': [['IFLA_XDP_ATTACHED', None]]}],
                (
                    'IFLA_PERM_ADDRESS',
                    self.perm_address if self.perm_address else self.address,
                ),
                [
                    'IFLA_AF_SPEC',
                    {
                        'attrs': [
                            [
                                'AF_INET',
                                {
                                    'accept_local': 0,
                                    'accept_redirects': 1,
                                    'accept_source_route': 0,
                                    'arp_accept': 0,
                                    'arp_announce': 0,
                                    'arp_ignore': 0,
                                    'arp_notify': 0,
                                    'arpfilter': 0,
                                    'bootp_relay': 0,
                                    'dummy': 65672,
                                    'force_igmp_version': 0,
                                    'forwarding': 1,
                                    'log_martians': 0,
                                    'mc_forwarding': 0,
                                    'medium_id': 0,
                                    'nopolicy': 1,
                                    'noxfrm': 1,
                                    'promote_secondaries': 1,
                                    'proxy_arp': 0,
                                    'proxy_arp_pvlan': 0,
                                    'route_localnet': 0,
                                    'rp_filter': 2,
                                    'secure_redirects': 1,
                                    'send_redirects': 1,
                                    'shared_media': 1,
                                    'src_vmark': 0,
                                    'tag': 0,
                                },
                            ]
                        ]
                    },
                ],
            ],
            'change': 0,
            'event': 'RTM_NEWLINK',
            'family': 0,
            'flags': self.flags,
            'header': {
                'error': None,
                'flags': 2,
                'length': 1364,
                'pid': 303471,
                'sequence_number': 260,
                'stats': Stats(qsize=0, delta=0, delay=0),
                'target': 'localhost',
                'type': 16,
            },
            'ifi_type': 772,
            'index': self.index,
            'state': 'up' if self.flags & 1 else 'down',
        }
        linkinfo = None
        infodata = {'attrs': []}
        if self.kind is not None:
            linkinfo = {'attrs': [('IFLA_INFO_KIND', self.kind)]}
        if self.kind not in (None, 'dummy'):
            linkinfo['attrs'].append(('IFLA_INFO_DATA', infodata))
        if self.kind == 'vlan':
            infodata['attrs'].append(('IFLA_VLAN_ID', self.vlan_id))
            ret['attrs'].append(('IFLA_LINK', self.link))
        if self.kind == 'bridge':
            infodata['attrs'].extend(
                (
                    ('IFLA_BR_MAX_AGE', self.br_max_age),
                    ('IFLA_BR_FORWARD_DELAY', self.br_forward_delay),
                )
            )
        if linkinfo is not None:
            ret['attrs'].append(('IFLA_LINKINFO', linkinfo))
        if self.master != 0:
            ret['attrs'].append(('IFLA_MASTER', self.master))
        return ret


class MockAddress:
    def __init__(
        self,
        index=0,
        address=None,
        prefixlen=None,
        broadcast=None,
        label=None,
        family=2,
        local=None,
        **kwarg,
    ):
        self.address = address
        self.local = local
        self.broadcast = broadcast
        self.prefixlen = prefixlen
        self.index = index
        self.label = label
        self.family = family

    def update_from_msg(self, msg):
        [
            setattr(self, x, msg.get(x))
            for x in ['index', 'address', 'broadcast', 'prefixlen']
            if msg.get(x) is not None
        ]

    @classmethod
    def load_from_msg(cls, msg):
        ret = cls()
        ret.update_from_msg(msg)
        return ret

    def export(self):
        ret = {
            'family': self.family,
            'prefixlen': self.prefixlen,
            'flags': 0,
            'scope': 0,
            'index': self.index,
            'attrs': [
                ('IFA_ADDRESS', self.address),
                ('IFA_LOCAL', self.local if self.local else self.address),
                ('IFA_FLAGS', 512),
                (
                    'IFA_CACHEINFO',
                    {
                        'ifa_preferred': 3476,
                        'ifa_valid': 3476,
                        'cstamp': 138655779,
                        'tstamp': 141288674,
                    },
                ),
            ],
            'header': {
                'length': 88,
                'type': 20,
                'flags': 2,
                'sequence_number': 256,
                'pid': 320994,
                'error': None,
                'target': 'localhost',
                'stats': Stats(qsize=0, delta=0, delay=0),
            },
            'event': 'RTM_NEWADDR',
        }
        if self.label is not None:
            ret['attrs'].append(('IFA_LABEL', self.label))
        if self.broadcast is not None:
            ret['attrs'].append(('IFA_BROADCAST', self.broadcast))
        return ret


class MockRoute:
    def __init__(
        self,
        dst=None,
        oif=0,
        gateway=None,
        prefsrc=None,
        family=2,
        dst_len=24,
        table=254,
        scope=253,
        proto=2,
        route_type=1,
        **kwarg,
    ):
        self.dst = dst
        self.gateway = gateway
        self.prefsrc = prefsrc
        self.oif = oif
        self.family = family
        self.dst_len = dst_len
        self.table = table
        self.scope = scope
        self.proto = proto
        self.route_type = route_type
        self.priority = kwarg.get('priority', 0)
        self.tos = kwarg.get('tos', 0)
        self._type = kwarg.get('type', 2)

    def update_from_msg(self, msg):
        [
            setattr(self, x, msg.get(x))
            for x in [
                'dst',
                'dst_len',
                'gateway',
                'oif',
                'family',
                'table',
                'priority',
            ]
            if msg.get(x) is not None
        ]

    @classmethod
    def load_from_msg(cls, msg):
        ret = cls()
        ret.update_from_msg(msg)
        return ret

    def export(self):
        ret = {
            'family': self.family,
            'dst_len': self.dst_len,
            'src_len': 0,
            'tos': self.tos,
            'table': self.table if self.table <= 255 else 252,
            'proto': self.proto,
            'scope': self.scope,
            'type': self._type,
            'flags': 0,
            'attrs': [('RTA_TABLE', self.table), ('RTA_OIF', self.oif)],
            'header': {
                'length': 60,
                'type': 24,
                'flags': 2,
                'sequence_number': 255,
                'pid': 325359,
                'error': None,
                'target': 'localhost',
                'stats': Stats(qsize=0, delta=0, delay=0),
            },
            'event': 'RTM_NEWROUTE',
        }
        if self.dst is not None:
            ret['attrs'].append(('RTA_DST', self.dst))
        if self.prefsrc is not None:
            ret['attrs'].append(('RTA_PREFSRC', self.prefsrc))
        if self.gateway is not None:
            ret['attrs'].append(('RTA_GATEWAY', self.gateway))
        if self.priority > 0:
            ret['attrs'].append(('RTA_PRIORITY', self.priority))
        return ret


presets = {
    'default': {
        'links': [
            MockLink(
                index=1,
                ifname='lo',
                address='00:00:00:00:00:00',
                broadcast='00:00:00:00:00:00',
                rx_bytes=43309665,
                tx_bytes=43309665,
                rx_packets=173776,
                tx_packets=173776,
                mtu=65536,
                qdisc='noqueue',
            ),
            MockLink(
                index=2,
                ifname='eth0',
                address='52:54:00:72:58:b2',
                broadcast='ff:ff:ff:ff:ff:ff',
                rx_bytes=175340,
                tx_bytes=175340,
                rx_packets=10251,
                tx_packets=10251,
                mtu=1500,
                qdisc='fq_codel',
            ),
        ],
        'addr': [
            MockAddress(
                index=1,
                label='lo',
                address='127.0.0.1',
                broadcast='127.255.255.255',
                prefixlen=8,
            ),
            MockAddress(
                index=2,
                label='eth0',
                address='192.168.122.28',
                broadcast='192.168.122.255',
                prefixlen=24,
            ),
        ],
        'routes': [
            MockRoute(
                dst=None,
                gateway='192.168.122.1',
                oif=2,
                dst_len=0,
                table=254,
                scope=0,
            ),
            MockRoute(dst='192.168.122.0', oif=2, dst_len=24, table=254),
            MockRoute(
                dst='127.0.0.0', oif=1, dst_len=8, table=255, route_type=2
            ),
            MockRoute(
                dst='127.0.0.1', oif=1, dst_len=32, table=255, route_type=2
            ),
            MockRoute(
                dst='127.255.255.255',
                oif=1,
                dst_len=32,
                table=255,
                route_type=3,
            ),
            MockRoute(
                dst='192.168.122.28',
                oif=2,
                dst_len=32,
                table=255,
                route_type=2,
            ),
            MockRoute(
                dst='192.168.122.255',
                oif=2,
                dst_len=32,
                table=255,
                route_type=3,
            ),
        ],
    },
    'netns': {
        'links': [
            MockLink(
                index=1,
                ifname='lo',
                address='00:00:00:00:00:00',
                broadcast='00:00:00:00:00:00',
                rx_bytes=43309665,
                tx_bytes=43309665,
                rx_packets=173776,
                tx_packets=173776,
                mtu=65536,
                qdisc='noqueue',
            )
        ],
        'addr': [
            MockAddress(
                index=1,
                label='lo',
                address='127.0.0.1',
                broadcast='127.255.255.255',
                prefixlen=8,
            )
        ],
        'routes': [
            MockRoute(
                dst='127.0.0.0', oif=1, dst_len=8, table=255, route_type=2
            ),
            MockRoute(
                dst='127.0.0.1', oif=1, dst_len=32, table=255, route_type=2
            ),
            MockRoute(
                dst='127.255.255.255',
                oif=1,
                dst_len=32,
                table=255,
                route_type=3,
            ),
        ],
    },
}


class IPEngine:
    '''Mock network objects database with the socket API.

    WIP: work in progress.

    A drop-in replacement to use instead of a low level RTNL socket.
    Implements all the required socket properties and provides a
    network objects database with RTNL protocol.

    Example::

        >>> ipe = IPEngine()
        >>> ipr = IPRoute(use_socket=ipe)
        >>> [ x.get('ifname') for x in ipr.link('dump') ]
        ['lo', 'eth0']

    '''

    def __init__(
        self,
        sfamily=AF_NETLINK,
        stype=socket.SOCK_DGRAM,
        sproto=0,
        netns='default',
        flags=os.O_CREAT,
    ):
        self.marshal = MarshalRtnl()
        self.netns = netns
        self.flags = flags
        self.magic = 0
        self._stype = stype
        self._sfamily = sfamily
        self._sproto = sproto
        self._local = threading.local()
        self._lock = threading.Lock()
        self._broadcast = set()
        self.processors = {
            RTM_GETADDR: self.RTM_GETADDR,
            RTM_GETLINK: self.RTM_GETLINK,
            RTM_NEWADDR: self.RTM_NEWADDR,
            RTM_DELADDR: self.RTM_DELADDR,
            RTM_NEWLINK: self.RTM_NEWLINK,
            RTM_DELROUTE: self.RTM_DELROUTE,
            RTM_NEWROUTE: self.RTM_NEWROUTE,
            RTM_GETROUTE: self.RTM_GETROUTE,
        }
        self.initdb()

    @property
    def loopback_r(self):
        if not hasattr(self._local, 'loopback_r'):
            self._local.loopback_r, self._local.loopback_w = socket.socketpair(
                socket.AF_UNIX, socket.SOCK_DGRAM
            )
        return self._local.loopback_r

    @property
    def loopback_w(self):
        if not hasattr(self._local, 'loopback_w'):
            self._local.loopback_r, self._local.loopback_w = socket.socketpair(
                socket.AF_UNIX, socket.SOCK_DGRAM
            )
        return self._local.loopback_w

    def initdb(self):
        if self.netns not in presets:
            if not self.flags & os.O_CREAT:
                raise FileNotFoundError()
            presets[self.netns] = copy.deepcopy(presets['netns'])
        self.database = copy.deepcopy(presets[self.netns])

    def close(self):
        if self.magic == MAGIC_CLOSE:
            self.loopback_r.close()
            self.loopback_w.close()

    def bind(self, address=None):
        self._broadcast.add(self.loopback_w)
        msg = rtmsg()
        msg.load(self.database['routes'][0].export())
        msg.encode()
        self.loopback_w.send(msg.data)

    def fileno(self):
        return self.loopback_r.fileno()

    def recv(self, bufsize, flags=0):
        return self.loopback_r.recv(bufsize, flags)

    def recvfrom(self, bufsize, flags=0):
        return self.loopback_r.recvfrom(bufsize, flags)

    def recvmsg(self, bufsize, ancbufsize=0, flags=0):
        return self.loopback_r.recvmsg(bufsize, ancbufsize, flags)

    def recv_into(self, buffer, nbytes=0, flags=0):
        return self.loopback_r.recv_into(buffer, nbytes, flags)

    def recvfrom_into(self, buffer, nbytes=0, flags=0):
        return self.loopback_r.recvfrom_into(buffer, nbytes, flags)

    def recvmsg_into(self, buffers, ancbufsize=0, flags=0):
        return self.loopback_r.recvmsg_into(buffers, ancbufsize, flags)

    def send(self, data, flags=0):
        return self.nl_handle(data)

    def sendall(self, data, flags=0):
        return self.nl_handle(data)

    def sendto(self, data, flags, address=0):
        return self.nl_handle(data)

    def sendmsg(self, buffers, ancdata=None, flags=None, address=None):
        raise NotImplementedError()

    def setblocking(self, flag):
        return self.loopback_r.setblocking(flag)

    def getblocking(self):
        return self.loopback_r.getblocking()

    def getsockname(self):
        return self.loopback_r.getsockname()

    def getpeername(self):
        return self.loopback_r.getpeername()

    @property
    def type(self):
        return self._stype

    @property
    def family(self):
        return self._sfamily

    @property
    def proto(self):
        return self._sproto

    def nl_handle(self, data):
        with self._lock:
            for msg in self.marshal.parse(data):
                key = msg['header']['type']
                tag = msg['header']['sequence_number']
                if key in self.processors:
                    self.processors[key](msg)
                else:
                    self.nl_done(tag)
            return len(data)

    def nl_dump(self, registry, msg_class, tag):
        for item in registry:
            msg = msg_class()
            msg.load(item.export())
            msg['header']['flags'] = NLM_F_MULTI
            msg['header']['sequence_number'] = tag
            msg.encode()
            yield msg

    def nl_broadcast(self, msg):
        for sock in self._broadcast:
            msg['header']['sequence_number'] = 0
            msg.reset()
            msg.encode()
            sock.send(msg.data)

    def nl_done(self, tag):
        msg = nlmsg()
        msg['header']['type'] = NLMSG_DONE
        msg['header']['sequence_number'] = tag
        msg.encode()
        self.loopback_w.send(msg.data)

    def nl_error(self, tag, code):
        msg = nlmsgerr()
        msg['header']['sequence_number'] = tag
        msg['error'] = code
        msg.encode()
        self.loopback_w.send(msg.data)

    def RTM_GETROUTE(self, req):
        tag = req['header']['sequence_number']
        for msg in self.nl_dump(self.database['routes'], rtmsg, tag):
            self.loopback_w.send(msg.data)
        self.nl_done(tag)

    def RTM_GETADDR(self, msg_in):
        tag = msg_in['header']['sequence_number']
        for msg_out in self.nl_dump(self.database['addr'], ifaddrmsg, tag):
            self.loopback_w.send(msg_out.data)
        self.nl_done(tag)

    def RTM_GETLINK(self, msg_in):
        tag = msg_in['header']['sequence_number']
        database = self.database['links']
        if msg_in.get('index') > 0:
            database = [
                x
                for x in self.database['links']
                if x.index == msg_in.get('index')
            ]
        elif msg_in.get('ifname'):
            database = [
                x
                for x in self.database['links']
                if x.ifname == msg_in.get('ifname')
            ]
        for msg_out in self.nl_dump(database, ifinfmsg, tag):
            self.loopback_w.send(msg_out.data)
        if msg_in.get(('header', 'flags')) & NLM_F_DUMP:
            return self.nl_done(tag)
        self.nl_error(tag, 0)

    def RTM_NEWROUTE(self, req):
        tag = req['header']['sequence_number']
        if not req.get('oif'):
            (gateway,) = struct.unpack(
                '>I', socket.inet_aton(req.get('gateway'))
            )
            for route in self.database['routes']:
                if route.dst is None:
                    continue
                (dst,) = struct.unpack('>I', socket.inet_aton(route.dst))
                if (gateway & (0xFFFFFFFF << (32 - route.dst_len))) == dst:
                    req['attrs'].append(('RTA_OIF', route.oif))
                    break
            else:
                return self.nl_error(tag, errno.ENOENT)
        idx = {
            (x.dst, x.dst_len, x.oif, x.priority, x.gateway, x.table): x
            for x in self.database['routes']
        }
        req_index = (
            req.get('dst'),
            req.get('dst_len'),
            req.get('oif'),
            req.get('priority'),
            req.get('gateway'),
            req.get('IFLA_TABLE') or req.get('table'),
        )
        if req_index in idx:
            return self.nl_error(tag, errno.EEXIST)
        route = MockRoute.load_from_msg(req)
        self.database['routes'].append(route)
        self.nl_error(tag, 0)
        msg = rtmsg()
        msg.load(route.export())
        self.nl_broadcast(msg)

    def RTM_DELROUTE(self, req):
        idx = {
            (x.dst, x.dst_len, x.oif, x.priority, x.table): x
            for x in self.database['routes']
        }
        req_index = (
            req.get('dst'),
            req.get('dst_len'),
            req.get('oif'),
            req.get('priority'),
            req.get('IFLA_TABLE') or req.get('table'),
        )
        tag = req['header']['sequence_number']
        if req_index not in idx:
            return self.nl_error(tag, errno.ENOENT)
        self.database['routes'].remove(idx[req_index])
        self.nl_error(tag, 0)
        self.nl_broadcast(req)

    def RTM_DELADDR(self, req):
        idx = {
            (x.index, x.address, x.prefixlen): x for x in self.database['addr']
        }
        req_index = (
            req.get("index"),
            req.get("address"),
            req.get("prefixlen"),
        )
        tag = req['header']['sequence_number']
        if req_index not in idx:
            return self.nl_error(tag, errno.ENOENT)
        self.database['addr'].remove(idx[req_index])
        self.nl_error(tag, 0)
        self.nl_broadcast(req)

    def RTM_NEWADDR(self, req):
        idx = {
            (x.index, x.address, x.prefixlen) for x in self.database['addr']
        }
        req_index = (
            req.get("index"),
            req.get("address"),
            req.get("prefixlen"),
        )
        tag = req['header']['sequence_number']
        if req_index in idx:
            return self.nl_error(tag, errno.EEXIST)
        addr = MockAddress.load_from_msg(req)
        self.database['addr'].append(addr)
        self.nl_error(tag, 0)
        msg = ifaddrmsg()
        msg.load(addr.export())
        self.nl_broadcast(msg)

    def RTM_NEWLINK(self, req):
        idx = {x.index: x for x in self.database['links']}
        nmx = {x.ifname: x for x in self.database['links']}
        tag = req['header']['sequence_number']
        if req.get('index') in idx:
            link = idx[req.get('index')]
            if link.ifname == req.get('ifname'):
                return self.nl_error(tag, errno.EEXIST)
            link.update_from_msg(req)
        elif req.get('ifname') in nmx:
            return self.nl_error(tag, errno.EEXIST)
        else:
            link = MockLink.load_from_msg(req)
            self.database['links'].append(link)
        self.nl_error(tag, 0)
        msg = ifinfmsg()
        msg.load(link.export())
        self.nl_broadcast(msg)



import pytest

from pyroute2.netlink.buffer import Buffer, Page

buffer_settings = (
    'mode,size,page_size',
    (('internal', 10485760, 32768), ('shared', 10485760, 32768)),
)


@pytest.mark.parametrize(*buffer_settings)
def test_create_buffer(mode, size, page_size):
    try:
        buffer = Buffer(mode, size, page_size)
    except ModuleNotFoundError:
        pytest.skip(f'buffer mode "{mode}" not supported')
    assert buffer.mode == mode
    assert buffer.size == size
    assert buffer.page_size == page_size
    minimal_index = 0
    maximal_index = size // page_size
    assert len(buffer.directory) == maximal_index
    for index, page in buffer.directory.items():
        assert minimal_index <= index <= maximal_index
        assert isinstance(page, Page)
        assert page.offset == index * page_size
        assert page.is_free is True
    buffer.close()


@pytest.mark.parametrize(*buffer_settings)
def test_use_all_pages(mode, size, page_size):
    try:
        buffer = Buffer(mode, size, page_size)
    except ModuleNotFoundError:
        pytest.skip(f'buffer mode "{mode}" not supported')
    maximal_index = size // page_size
    marker = 0x05
    for _ in range(maximal_index):
        page = buffer.get_free_page()
        assert not page.is_free
        page.view[0] = marker
        assert page.view[0] == marker
        assert buffer.view[page.offset] == marker
        assert buffer.buf[page.offset] == marker
        marker += 1
        if marker == 0xFF:
            marker = 0x05

    with pytest.raises(MemoryError):
        buffer.get_free_page()
    buffer.close()


@pytest.mark.parametrize(*buffer_settings)
def test_context_manager(mode, size, page_size):
    try:
        with Buffer(mode, size, page_size) as buffer:
            assert buffer.mode == mode
    except ModuleNotFoundError:
        pytest.skip(f'buffer mode "{mode}" not supported')



from pyroute2.netlink import nla


class gtp(nla):
    nla_map = (
        ('IFLA_GTP_UNSPEC', 'none'),
        ('IFLA_GTP_FD0', 'uint32'),
        ('IFLA_GTP_FD1', 'uint32'),
        ('IFLA_GTP_PDP_HASHSIZE', 'uint32'),
    )



"""
nf_tables expression netlink attributes

See EXPRESSIONS in nft(8).
"""

from socket import AF_INET, AF_INET6

from pyroute2.nftables.parser.parser import conv_map_tuple, nfta_nla_parser


class NFTReg(object):
    def __init__(self, num):
        self.num = num

    @classmethod
    def from_netlink(cls, nlval):
        # please, for more information read nf_tables.h.
        if nlval == 'NFT_REG_VERDICT':
            num = 0
        else:
            num = int(nlval.split('_')[-1].lower())
            if nlval.startswith('NFT_REG32_'):
                num += 8
        return cls(num=num)

    @staticmethod
    def to_netlink(reg):
        # please, for more information read nf_tables.h.
        if reg.num == 0:
            return 'NFT_REG_VERDICT'
        if reg.num < 8:
            return 'NFT_REG_{0}'.format(reg.num)
        return 'NFT_REG32_{0}'.format(reg.num)

    @classmethod
    def from_dict(cls, val):
        return cls(num=val)

    def to_dict(self):
        return self.num


class NFTVerdict(object):
    def __init__(self, verdict, chain):
        self.verdict = verdict
        self.chain = chain

    @classmethod
    def from_netlink(cls, ndmsg):
        if ndmsg.get_attr('NFTA_VERDICT_CODE') is not None:
            verdict = (
                ndmsg.get_attr('NFTA_VERDICT_CODE').split('_')[-1].lower()
            )
        else:
            verdict = None
        chain = ndmsg.get_attr('NFTA_VERDICT_CHAIN')
        return cls(verdict=verdict, chain=chain)

    def to_netlink(self):
        attrs = [('NFTA_VERDICT_CODE', 'NF_' + self.verdict)]
        if self.chain is not None:
            attrs.append(('NFTA_VERDICT_CHAIN', self.chain))
        return attrs

    @classmethod
    def from_dict(cls, d):
        return cls(verdict=d['verdict'], chain=d.get('chain', None))

    def to_dict(self):
        d = {'verdict': self.verdict}
        if self.chain is not None:
            d['chain'] = self.chain
        return d


class NFTData(object):
    def __init__(self, data_type, data):
        self.type = data_type
        self.data = data

    def to_netlink(self):
        if self.type == 'value':
            return ('NFTA_DATA_VALUE', self.data)
        if self.type == 'verdict':
            return ('NFTA_DATA_VERDICT', self.data)
        raise NotImplementedError(self.type)

    @classmethod
    def from_netlink(cls, ndmsg):
        if ndmsg.get_attr('NFTA_DATA_VALUE') is not None:
            kwargs = {
                'data_type': 'value',
                'data': ndmsg.get_attr('NFTA_DATA_VALUE'),
            }
        elif ndmsg.get_attr('NFTA_DATA_VERDICT') is not None:
            kwargs = {
                'data_type': 'verdict',
                'data': NFTVerdict.from_netlink(
                    ndmsg.get_attr('NFTA_DATA_VERDICT')
                ),
            }
        else:
            raise NotImplementedError(ndmsg)
        return cls(**kwargs)

    @classmethod
    def from_dict(cls, data):
        def from_32hex(val):
            return bytes(bytearray.fromhex(val[2:]))[::-1]

        kwargs = {}
        data = data["reg"]
        if data['type'] == 'value':
            value = bytes()
            for i in range(0, data['len'], 4):
                value += from_32hex(data['data{0}'.format(i / 4)])
            kwargs['data'] = value
        elif data['type'] == 'verdict':
            kwargs['data'] = NFTVerdict.from_dict(data)
        else:
            raise NotImplementedError()
        kwargs['data_type'] = data['type']
        return cls(**kwargs)

    def to_dict(self):
        def to_32hex(s):
            res = ''.join('{:02x}'.format(c) for c in bytearray(s)[::-1])
            return '0x' + res.zfill(8)

        if self.type == 'value':
            len_data = len(self.data)
            d = {'type': 'value', 'len': len_data}
            for i in range(0, len_data, 4):
                d['data{0}'.format(i / 4)] = to_32hex(self.data[i : i + 4])
        elif self.type == 'verdict':
            d = self.data.to_dict()
            d['type'] = 'verdict'
        else:
            raise NotImplementedError()
        return {"reg": d}


class NFTRuleExpr(nfta_nla_parser):
    #######################################################################
    conv_maps = (conv_map_tuple('name', 'NFTA_EXPR_NAME', 'type', 'raw'),)
    #######################################################################

    @classmethod
    def from_netlink(cls, expr_type, ndmsg):
        inst = super(NFTRuleExpr, cls).from_netlink(ndmsg)
        inst.name = expr_type
        return inst

    cparser_reg = NFTReg
    cparser_data = NFTData

    class cparser_extract_str(object):
        STRVAL = None

        @classmethod
        def from_netlink(cls, val):
            magic = '{0}'
            left, right = cls.STRVAL.split(magic, 1)
            if right:
                val = val[len(left) : -len(right)]
            else:
                val = val[len(left) :]
            return val.lower()

        @classmethod
        def to_netlink(cls, val):
            return cls.STRVAL.format(val).upper()

        @staticmethod
        def from_dict(val):
            return val

        @staticmethod
        def to_dict(val):
            return val

    class cparser_inet_family(object):
        @staticmethod
        def from_netlink(val):
            if val == AF_INET:
                return 'ip'
            if val == AF_INET6:
                return 'ip6'
            return val

        @staticmethod
        def to_netlink(val):
            if val == 'ip':
                return AF_INET
            if val == 'ip6':
                return AF_INET6
            return val

        @staticmethod
        def from_dict(val):
            return val

        @staticmethod
        def to_dict(val):
            return val


class ExprMeta(NFTRuleExpr):
    conv_maps = NFTRuleExpr.conv_maps + (
        conv_map_tuple('key', 'NFTA_META_KEY', 'key', 'meta_key'),
        conv_map_tuple('dreg', 'NFTA_META_DREG', 'dreg', 'reg'),
    )

    class cparser_meta_key(NFTRuleExpr.cparser_extract_str):
        STRVAL = 'NFT_META_{0}'


class ExprCmp(NFTRuleExpr):
    conv_maps = NFTRuleExpr.conv_maps + (
        conv_map_tuple('sreg', 'NFTA_CMP_SREG', 'sreg', 'reg'),
        conv_map_tuple('op', 'NFTA_CMP_OP', 'op', 'cmp_op'),
        conv_map_tuple('data', 'NFTA_CMP_DATA', 'data', 'data'),
    )

    class cparser_cmp_op(NFTRuleExpr.cparser_extract_str):
        STRVAL = 'NFT_CMP_{0}'


class ExprImmediate(NFTRuleExpr):
    conv_maps = NFTRuleExpr.conv_maps + (
        conv_map_tuple('dreg', 'NFTA_IMMEDIATE_DREG', 'dreg', 'reg'),
        conv_map_tuple('data', 'NFTA_IMMEDIATE_DATA', 'data', 'data'),
    )


class ExprPayload(NFTRuleExpr):
    conv_maps = NFTRuleExpr.conv_maps + (
        conv_map_tuple('dreg', 'NFTA_PAYLOAD_DREG', 'dreg', 'reg'),
        conv_map_tuple('base', 'NFTA_PAYLOAD_BASE', 'base', 'payload_base'),
        conv_map_tuple('offset', 'NFTA_PAYLOAD_OFFSET', 'offset', 'raw'),
        conv_map_tuple('len', 'NFTA_PAYLOAD_LEN', 'len', 'raw'),
    )

    class cparser_payload_base(NFTRuleExpr.cparser_extract_str):
        STRVAL = 'NFT_PAYLOAD_{0}_HEADER'

        @classmethod
        def from_netlink(cls, ndmsg):
            val = super(ExprPayload.cparser_payload_base, cls).from_netlink(
                ndmsg
            )
            if val == 'll':
                return 'link'
            return val

        @classmethod
        def to_netlink(cls, val):
            if val == 'link':
                val = 'll'
            return super(ExprPayload.cparser_payload_base, cls).to_netlink(val)

        @staticmethod
        def from_dict(val):
            return val

        @staticmethod
        def to_dict(val):
            return val


class ExprLookup(NFTRuleExpr):
    conv_maps = NFTRuleExpr.conv_maps + (
        conv_map_tuple('setid', 'NFTA_LOOKUP_SET', 'set', 'raw'),
        conv_map_tuple('sreg', 'NFTA_LOOKUP_SREG', 'sreg', 'reg'),
        conv_map_tuple('flags', 'NFTA_LOOKUP_FLAGS', 'flags', 'raw'),
    )


class ExprNat(NFTRuleExpr):
    conv_maps = NFTRuleExpr.conv_maps + (
        conv_map_tuple('nat_type', 'NFTA_NAT_TYPE', 'nat_type', 'nat_type'),
        conv_map_tuple('family', 'NFTA_NAT_FAMILY', 'family', 'inet_family'),
        conv_map_tuple(
            'sreg_addr_min', 'NFTA_NAT_REG_ADDR_MIN', 'sreg_addr_min', 'reg'
        ),
        conv_map_tuple(
            'sreg_addr_max', 'NFTA_NAT_REG_ADDR_MAX', 'sreg_addr_max', 'reg'
        ),
    )

    class cparser_nat_type(NFTRuleExpr.cparser_extract_str):
        STRVAL = 'NFT_NAT_{0}'


class ExprBitwise(NFTRuleExpr):
    conv_maps = NFTRuleExpr.conv_maps + (
        conv_map_tuple('sreg', 'NFTA_BITWISE_SREG', 'sreg', 'reg'),
        conv_map_tuple('dreg', 'NFTA_BITWISE_DREG', 'dreg', 'reg'),
        conv_map_tuple('len', 'NFTA_BITWISE_LEN', 'len', 'raw'),
        conv_map_tuple('mask', 'NFTA_BITWISE_MASK', 'mask', 'data'),
        conv_map_tuple('xor', 'NFTA_BITWISE_XOR', 'xor', 'data'),
    )


class ExprCounter(NFTRuleExpr):
    conv_maps = NFTRuleExpr.conv_maps + (
        conv_map_tuple('bytes', 'NFTA_COUNTER_BYTES', 'bytes', 'raw'),
        conv_map_tuple('packets', 'NFTA_COUNTER_PACKETS', 'pkts', 'raw'),
    )


class ExprMatch(NFTRuleExpr):
    conv_maps = NFTRuleExpr.conv_maps + (
        conv_map_tuple('match_name', 'NFTA_MATCH_NAME', 'name', 'raw'),
    )


class ExprTarget(NFTRuleExpr):
    conv_maps = NFTRuleExpr.conv_maps + (
        conv_map_tuple('target_name', 'NFTA_TARGET_NAME', 'name', 'raw'),
    )


NFTA_EXPR_NAME_MAP = {
    'meta': ExprMeta,
    'cmp': ExprCmp,
    'immediate': ExprImmediate,
    'payload': ExprPayload,
    'lookup': ExprLookup,
    'nat': ExprNat,
    'bitwise': ExprBitwise,
    'counter': ExprCounter,
    'match': ExprMatch,
    'target': ExprTarget,
}


def get_expression_from_netlink(ndmsg):
    name = ndmsg.get_attr('NFTA_EXPR_NAME')
    try:
        expr_cls = NFTA_EXPR_NAME_MAP[name]
    except KeyError:
        raise NotImplementedError(
            "can't load rule expression {0} from netlink {1}".format(
                name, ndmsg
            )
        )
    return expr_cls.from_netlink(name, ndmsg.get_attr('NFTA_EXPR_DATA'))


def get_expression_from_dict(d):
    name = d['type']
    if name in NFTA_EXPR_NAME_MAP:
        expr_cls = NFTA_EXPR_NAME_MAP[name]
    else:
        raise NotImplementedError(
            "can't load rule expression {0} from json {1}".format(name, d)
        )
    return expr_cls.from_dict(d)



import pytest

from pyroute2.common import AddrPool


def test_alloc_aligned():
    ap = AddrPool(minaddr=1, maxaddr=1024)
    for i in range(1024):
        ap.alloc()
    with pytest.raises(KeyError):
        ap.alloc()


def test_alloc_odd():
    ap = AddrPool(minaddr=1, maxaddr=1020)
    for i in range(1020):
        ap.alloc()
    with pytest.raises(KeyError):
        ap.alloc()


def test_reverse():
    ap = AddrPool(minaddr=1, maxaddr=1024, reverse=True)
    for i in range(512):
        assert ap.alloc() > ap.alloc()


def test_free():
    ap = AddrPool(minaddr=1, maxaddr=1024)
    f = ap.alloc()
    ap.free(f)


def test_free_fail():
    ap = AddrPool(minaddr=1, maxaddr=1024)
    with pytest.raises(KeyError):
        ap.free(0)


def test_free_reverse_fail():
    ap = AddrPool(minaddr=1, maxaddr=1024, reverse=True)
    with pytest.raises(KeyError):
        ap.free(0)


def test_locate():
    ap = AddrPool()
    f = ap.alloc()
    base1, bit1, is_allocated1 = ap.locate(f)
    base2, bit2, is_allocated2 = ap.locate(f + 1)
    assert base1 == base2
    assert bit2 == bit1 + 1
    assert is_allocated1
    assert not is_allocated2



import grp
import io
import json
import os
import pwd
import time
from dataclasses import dataclass
from functools import partial

from pyroute2.plan9 import Plan9Exit, Qid, Stat, Tcall, Tread, Twrite


def _publish_function_w(session, inode, request, response):
    inode.metadata.dirty = True
    inode.data.seek(0)
    inode.data.truncate()
    inode.data.write(request['data'])
    response['count'] = len(request['data'])
    return response


def _publish_function_r(
    func, loader, dumper, session, inode, request, response
):
    if request['offset'] == 0 and inode.metadata.has_new_data():
        try:
            kwarg = loader(inode.data.getvalue())
            ret = func(**kwarg)
            inode.metadata.dirty = False
        except Plan9Exit:
            raise
        except Exception as e:
            ret = e
        inode.data.seek(0)
        inode.data.truncate()
        inode.data.write(dumper(ret))

    inode.data.seek(request['offset'])
    response['data'] = inode.data.read(request['count'])
    return response


def _publish_function_c(
    func, loader, dumper, session, inode, request, response
):
    spec = loader(request['text'])

    if 'argv' not in spec:
        spec['argv'] = []
    if 'kwarg' not in spec:
        spec['kwarg'] = {}
    if 'data_arg' not in spec:
        spec['data_arg'] = 'data'
    data = request['data']
    if data:
        spec['kwarg'][spec['data_arg']] = data
    ret = func(*spec['argv'], **spec['kwarg'])

    response['err'] = 0
    response['text'] = ''
    response['data'] = dumper(ret)

    return response


@dataclass
class InodeMetadata:
    call_on_read: bool = False
    dirty: bool = False

    def has_new_data(self) -> bool:
        return self.dirty or self.call_on_read


class Inode:
    children = None
    parents = None
    qid = None
    stat = None
    data = None
    callbacks = None

    def __init__(
        self,
        name,
        path,
        data='',
        qtype=0,
        mode=0o600,
        uid=None,
        gid=None,
        parents=None,
        children=None,
    ):
        self.data = io.BytesIO(data.encode('utf-8'))
        self.parents = parents if parents is not None else set()
        self.children = children if children is not None else set()
        self.callbacks = {}
        self.metadata = InodeMetadata()
        self.stat = Stat()
        self.qid = Qid(qtype, 0, path)
        self.stat['uid'] = (
            uid if uid is not None else pwd.getpwuid(os.getuid()).pw_name
        )
        self.stat['gid'] = (
            gid if gid is not None else grp.getgrgid(os.getgid()).gr_name
        )
        self.stat['muid'] = self.stat['uid']
        self.stat['qid.type'] = self.qid['type']
        self.stat['qid.vers'] = self.qid['vers']
        self.stat['qid.path'] = self.qid['path']
        # shift qid.type 3 bytes left and OR with mode
        self.stat['mode'] = (self.qid['type'] << (8 * 3)) | mode
        self.stat['type'] = self.qid['path']
        self.stat['dev'] = 0
        self.stat['mtime'] = int(time.time())
        self.stat['atime'] = int(time.time())
        self.stat['name'] = name
        self.sync()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def sync(self):
        self.stat['length'] = len(self.data.getvalue())

    def get_parent(self, name=None):
        return tuple(self.parents)[0]

    def get_child(self, name):
        for child in self.children:
            if child.stat['name'] == name:
                return child
        raise KeyError('file not found')

    def register_function(
        self,
        func,
        loader=json.loads,
        dumper=lambda x: x if isinstance(x, bytes) else str(x).encode('utf-8'),
    ):
        self.add_callback(Twrite, _publish_function_w)
        self.add_callback(
            Tread, partial(_publish_function_r, func, loader, dumper)
        )
        self.add_callback(
            Tcall, partial(_publish_function_c, func, loader, dumper)
        )

    def add_callback(self, call, f):
        self.callbacks[call] = f
        return self

    def add_parent(self, inode):
        return self.parents.add(inode)

    def add_child(self, inode):
        inode.add_parent(self)
        return self.children.add(inode)

    def del_parent(self, inode, force=False):
        if force or len(self.parents) > 1:
            return self.parents.remove(inode)
        raise KeyError('can not remove the only parent')

    def del_child(self, inode):
        inode.del_parent(self, force=True)
        return self.children.remove(inode)

    def __id__(self):
        return self.qid['path']


class Filesystem:

    inodes = {}

    def __init__(self):
        self.__path = 255
        # create the root inode
        path = 0
        self.inodes[path] = root = Inode('/', path, qtype=0x80, mode=0o755)
        root.add_parent(root)

    @property
    def new_path(self):
        self.__path += 1
        return self.__path

    def walk(self, wname):
        inode = self.inodes[0]
        for name in wname.split(os.path.sep):
            if name == '':
                continue
            inode = inode.get_child(name)
        return inode

    def create(self, wname, qtype=0, mode=None, data=''):
        if mode is None:
            mode = 0o750 if qtype & 0x80 else 0o640
        pname, name = os.path.split(wname)
        parent = self.walk(pname)
        path = self.new_path
        inode = Inode(name, path, qtype=qtype, mode=mode, data=data)
        parent.add_child(inode)
        return inode

    def get_inode(self, path):
        return self.inodes[path]

    def set_inode(self, path, inode):
        self.inodes[path] = inode


class Session:

    fid_table = {}

    def __init__(self, filesystem):
        self.filesystem = filesystem

    @property
    def root(self):
        return self.filesystem.inodes[0]

    def set_fid(self, fid, inode):
        self.fid_table[fid] = inode

    def get_fid(self, fid):
        return self.fid_table[fid]

    def create(self, name, mode):
        pass



'''
Base netlink socket and marshal
===============================

All the netlink providers are derived from the socket
class, so they provide normal socket API, including
`getsockopt()`, `setsockopt()`, they can be used in
poll/select I/O loops etc.

asynchronous I/O
----------------

To run async reader thread, one should call
`NetlinkSocket.bind(async_cache=True)`. In that case
a background thread will be launched. The thread will
automatically collect all the messages and store
into a userspace buffer.

.. note::
    There is no need to turn on async I/O, if you
    don't plan to receive broadcast messages.

ENOBUF and async I/O
--------------------

When Netlink messages arrive faster than a program
reads then from the socket, the messages overflow
the socket buffer and one gets ENOBUF on `recv()`::

    ... self.recv(bufsize)
    error: [Errno 105] No buffer space available

One way to avoid ENOBUF, is to use async I/O. Then the
library not only reads and buffers all the messages, but
also re-prioritizes threads. Suppressing the parser
activity, the library increases the response delay, but
spares CPU to read and enqueue arriving messages as
fast, as it is possible.

With logging level DEBUG you can notice messages, that
the library started to calm down the parser thread::

    DEBUG:root:Packet burst: the reader thread priority
        is increased, beware of delays on netlink calls
        Counters: delta=25 qsize=25 delay=0.1

This state requires no immediate action, but just some
more attention. When the delay between messages on the
parser thread exceeds 1 second, DEBUG messages become
WARNING ones::

    WARNING:root:Packet burst: the reader thread priority
        is increased, beware of delays on netlink calls
        Counters: delta=2525 qsize=213536 delay=3

This state means, that almost all the CPU resources are
dedicated to the reader thread. It doesn't mean, that
the reader thread consumes 100% CPU -- it means, that the
CPU is reserved for the case of more intensive bursts. The
library will return to the normal state only when the
broadcast storm will be over, and then the CPU will be
100% loaded with the parser for some time, when it will
process all the messages queued so far.

when async I/O doesn't help
---------------------------

Sometimes, even turning async I/O doesn't fix ENOBUF.
Mostly it means, that in this particular case the Python
performance is not enough even to read and store the raw
data from the socket. There is no workaround for such
cases, except of using something *not* Python-based.

One can still play around with SO_RCVBUF socket option,
but it doesn't help much. So keep it in mind, and if you
expect massive broadcast Netlink storms, perform stress
testing prior to deploy a solution in the production.

classes
-------
'''

import asyncio
import errno
import logging
import os
import random
import struct
from dataclasses import dataclass
from socket import SO_RCVBUF, SO_SNDBUF, SOCK_DGRAM, SOL_SOCKET
from typing import Optional

from pyroute2 import config, netns
from pyroute2.common import AddrPool, basestring, msg_done
from pyroute2.config import AF_NETLINK
from pyroute2.netlink import (
    NETLINK_ADD_MEMBERSHIP,
    NETLINK_DROP_MEMBERSHIP,
    NETLINK_EXT_ACK,
    NETLINK_GENERIC,
    NETLINK_GET_STRICT_CHK,
    NETLINK_LISTEN_ALL_NSID,
    NLM_F_ACK,
    NLM_F_APPEND,
    NLM_F_ATOMIC,
    NLM_F_CREATE,
    NLM_F_DUMP,
    NLM_F_ECHO,
    NLM_F_EXCL,
    NLM_F_REPLACE,
    NLM_F_REQUEST,
    NLM_F_ROOT,
    SOL_NETLINK,
)
from pyroute2.netlink.core import (
    AsyncCoreSocket,
    CoreConfig,
    CoreDatagramProtocol,
    CoreSocketSpec,
    SyncAPI,
)
from pyroute2.netlink.exceptions import ChaoticException, NetlinkError
from pyroute2.netlink.marshal import Marshal
from pyroute2.netlink.rt_files import NlProtosFile
from pyroute2.requests.main import RequestFilter

log = logging.getLogger(__name__)


class CompileContext:
    def __init__(self, netlink_socket):
        self.netlink_socket = netlink_socket
        self.netlink_socket.compiled = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        self.netlink_socket.compiled = None


# 8<-----------------------------------------------------------
# Singleton, containing possible modifiers to the NetlinkSocket
# bind() call.
#
# Normally, you can open only one netlink connection for one
# process, but there is a hack. Current PID_MAX_LIMIT is 2^22,
# so we can use the rest to modify the pid field.
#
# See also libnl library, lib/socket.c:generate_local_port()
sockets = AddrPool(minaddr=0x0, maxaddr=0x3FF, reverse=True)
# 8<-----------------------------------------------------------


@dataclass
class NetlinkConfig(CoreConfig):
    family: int = NETLINK_GENERIC
    port: Optional[int] = None
    pid: Optional[int] = None
    fileno: Optional[int] = None
    rcvbuf: int = 1048576
    sndbuf: int = 1048576
    all_ns: bool = False
    ext_ack: bool = False
    strict_check: bool = False
    nlm_echo: bool = False


class NetlinkSocketSpecFilter(RequestFilter):
    def set_target(self, context, value):
        if 'target' in context:
            return {'target': context['target']}
        return {'target': value}

    def set_netns(self, context, value):
        if 'target' in context:
            return {'netns': value}
        return {'target': value, 'netns': value}

    def set_pid(self, context, value):
        if value is None:
            return {'pid': os.getpid() & 0x3FFFFF, 'port': context['port']}
        elif value == 0:
            return {'pid': os.getpid(), 'port': 0}
        else:
            return {'pid': value, 'port': 0}

    def set_port(self, context, value):
        if isinstance(value, int):
            return {'port': value, 'epid': context['pid'] + (value << 22)}


class NetlinkSocketSpec(CoreSocketSpec):
    defaults = {
        'pid': 0,
        'epid': 0,
        'port': 0,
        'closed': False,
        'uname': config.uname,
        'use_socket': False,
        'event_loop': 'none',
        'transport_mode': 'none',
    }
    status_filters = [NetlinkSocketSpecFilter]


class AsyncNetlinkSocket(AsyncCoreSocket):
    '''
    Netlink socket
    '''

    marshal_class = Marshal

    def __init__(
        self,
        family=NETLINK_GENERIC,
        port=None,
        pid=None,
        fileno=None,
        sndbuf=1048576,
        rcvbuf=1048576,
        rcvsize=16384,
        all_ns=False,
        async_qsize=None,
        nlm_generator=None,
        target='localhost',
        ext_ack=False,
        strict_check=False,
        groups=0,
        nlm_echo=False,
        use_socket=None,
        netns=None,
        flags=os.O_CREAT,
        libc=None,
        use_event_loop=None,
        telemetry=None,
    ):

        if isinstance(family, str):
            family = NlProtosFile().get_rt_id(family)

        # 8<-----------------------------------------
        self.spec = NetlinkSocketSpec(
            NetlinkConfig(
                family=family,
                port=port,
                pid=pid,
                fileno=fileno,
                sndbuf=sndbuf,
                rcvbuf=rcvbuf,
                rcvsize=rcvsize,
                all_ns=all_ns,
                target=target,
                ext_ack=ext_ack,
                strict_check=strict_check,
                groups=groups,
                nlm_echo=nlm_echo,
                tag_field='sequence_number',
                netns=netns,
                flags=flags,
                use_libc=libc is not None,
                use_socket=use_socket is not None,
                use_event_loop=use_event_loop is not None,
                telemetry=telemetry,
            )
        )
        # TODO: merge capabilities to self.status
        self.capabilities = {
            'create_bridge': config.kernel > [3, 2, 0],
            'create_bond': config.kernel > [3, 2, 0],
            'create_dummy': True,
            'provide_master': config.kernel[0] > 2,
        }
        super().__init__(
            libc=libc, use_socket=use_socket, use_event_loop=use_event_loop
        )
        self.marshal = self.marshal_class()
        self.request_proxy = None
        self.batch = None

    async def setup_endpoint(self):
        if getattr(self.local, 'transport', None) is not None:
            return
        self.local.transport, self.local.protocol = (
            await self.event_loop.create_datagram_endpoint(
                lambda: CoreDatagramProtocol(
                    self.connection_lost,
                    self.enqueue,
                    self._error_event,
                    self.status,
                ),
                sock=self.socket,
            )
        )

    def setup_socket(self):
        if self.status['use_socket']:
            return self.use_socket
        sock = netns.create_socket(
            netns=self.spec['netns'],
            family=AF_NETLINK,
            socket_type=SOCK_DGRAM,
            proto=self.spec['family'],  # netlink family = socket proto
            fileno=self.spec['fileno'],
            flags=self.spec['flags'],
            libc=self.libc,
        )
        sock.setsockopt(SOL_SOCKET, SO_SNDBUF, self.status['sndbuf'])
        sock.setsockopt(SOL_SOCKET, SO_RCVBUF, self.status['rcvbuf'])
        if self.status['ext_ack']:
            sock.setsockopt(SOL_NETLINK, NETLINK_EXT_ACK, 1)
        if self.status['all_ns']:
            sock.setsockopt(SOL_NETLINK, NETLINK_LISTEN_ALL_NSID, 1)
        if self.status['strict_check']:
            sock.setsockopt(SOL_NETLINK, NETLINK_GET_STRICT_CHK, 1)
        return sock

    @property
    def uname(self):
        return self.status['uname']

    @property
    def groups(self):
        return self.status['groups']

    @property
    def pid(self):
        return self.status['pid']

    @property
    def port(self):
        return self.status['port']

    @property
    def epid(self):
        return self.status['epid']

    @property
    def target(self):
        return self.status['target']

    def _sync_bind(self, groups=0, pid=None, **kwarg):
        self.spec['groups'] = groups
        # if we have pre-defined port, use it strictly
        self.spec['pid'] = pid
        if pid is None:
            for port in range(20, 200):
                try:
                    self.spec['port'] = port
                    self.socket.bind(
                        (self.status['epid'], self.status['groups'])
                    )
                    break
                except Exception as e:
                    # create a new underlying socket -- on kernel 4
                    # one failed bind() makes the socket useless
                    log.debug(e)
            else:
                raise KeyError('no free address available')

    async def bind(self, groups=0, pid=None, **kwarg):
        '''
        Bind the socket to given multicast groups, using
        given pid.

            - If pid is None, use automatic port allocation
            - If pid == 0, use process' pid
            - If pid == <int>, use the value instead of pid
        '''
        await self.setup_endpoint()
        self._sync_bind(groups, pid, **kwarg)

    def add_membership(self, group):
        self.socket.setsockopt(SOL_NETLINK, NETLINK_ADD_MEMBERSHIP, group)

    def drop_membership(self, group):
        self.socket.setsockopt(SOL_NETLINK, NETLINK_DROP_MEMBERSHIP, group)

    def enqueue(self, data, addr):
        # calculate msg_seq
        tag = struct.unpack_from('I', data, 8)[0]
        return self.msg_queue.put_nowait(tag, data)

    def compile(self):
        return CompileContext(self)

    def make_request_type(self, command, command_map):
        if isinstance(command, basestring):
            return (lambda x: (x[0], self.make_request_flags(x[1])))(
                command_map[command]
            )
        elif isinstance(command, int):
            return command, self.make_request_flags('create')
        elif isinstance(command, (list, tuple)):
            return command
        else:
            raise TypeError('allowed command types: int, str, list, tuple')

    def make_request_flags(self, mode):
        flags = {
            'dump': NLM_F_REQUEST | NLM_F_DUMP,
            'get': NLM_F_REQUEST | NLM_F_ACK,
            'req': NLM_F_REQUEST | NLM_F_ACK,
            'put': NLM_F_REQUEST | NLM_F_CREATE,
        }
        flags['create'] = flags['req'] | NLM_F_CREATE | NLM_F_EXCL
        flags['append'] = flags['req'] | NLM_F_CREATE | NLM_F_APPEND
        flags['change'] = flags['req'] | NLM_F_REPLACE
        flags['replace'] = flags['change'] | NLM_F_CREATE

        return flags[mode] | (
            NLM_F_ECHO
            if (self.status['nlm_echo'] and mode not in ('get', 'dump'))
            else 0
        )

    async def put(
        self,
        msg,
        msg_type,
        msg_flags=NLM_F_REQUEST,
        addr=(0, 0),
        msg_seq=0,
        msg_pid=None,
    ):
        request = NetlinkRequest(
            self,
            msg,
            msg_type=msg_type,
            msg_flags=msg_flags,
            msg_seq=msg_seq,
            msg_pid=msg_pid,
        )
        await request.send()
        return request

    async def nlm_request_batch(self, msgs, noraise=False):
        expected_responses = []
        data = b''
        for message in msgs:
            request = NetlinkRequest(self, message)
            await request.prepare()
            data += request.msg.data
            if (request.msg['header']['flags'] & NLM_F_ACK) or (
                request.msg['header']['flags'] & NLM_F_DUMP
            ):
                expected_responses.append(request)
            else:
                request.cleanup()

        self.send(data)
        for request in expected_responses:
            try:
                async for msg in request.response():
                    yield msg
            except NetlinkError:
                if not noraise:
                    raise

    async def nlm_request(
        self,
        msg,
        msg_type,
        msg_flags=NLM_F_REQUEST | NLM_F_DUMP,
        terminate=None,
        callback=None,
        parser=None,
    ):
        request = NetlinkRequest(
            self, msg, terminate=terminate, callback=callback
        )
        request.msg['header']['type'] = msg_type
        request.msg['header']['flags'] = msg_flags
        await request.send()
        return request.response()


class NetlinkRequest:
    # request flags
    flags = {
        'dump': NLM_F_REQUEST | NLM_F_DUMP,
        'root': NLM_F_REQUEST | NLM_F_ROOT | NLM_F_ATOMIC,
        'get': NLM_F_REQUEST | NLM_F_ACK,
        'req': NLM_F_REQUEST | NLM_F_ACK,
    }
    flags['create'] = flags['req'] | NLM_F_CREATE | NLM_F_EXCL
    flags['append'] = flags['req'] | NLM_F_CREATE | NLM_F_APPEND
    flags['change'] = flags['req'] | NLM_F_REPLACE
    flags['replace'] = flags['change'] | NLM_F_CREATE

    def __init__(
        self,
        sock,
        msg,
        command=None,
        command_map=None,
        dump_filter=None,
        request_filter=None,
        terminate=None,
        callback=None,
        parser=None,
        msg_type=None,
        msg_flags=None,
        msg_seq=None,
        msg_pid=None,
    ):
        self.sock = sock
        self.addr_pool = sock.addr_pool
        self.status = sock.status
        self.epid = sock.epid if msg_pid is None else msg_pid
        self.marshal = sock.marshal
        self.parser = parser
        # if not isinstance(msg, nlmsg):
        #    msg_class = self.marshal.msg_map[msg_type]
        #    msg = msg_class(msg)
        self.msg_seq = self.addr_pool.alloc() if msg_seq is None else msg_seq

        # prio 3: message object
        # prio 2: direct msg_type & msg_flags arguments
        # prio 1: command map
        if command_map is not None:
            msg_type, msg_flags = self.calculate_request_type(
                command, command_map, self.status['nlm_echo']
            )
        if msg_type is not None:
            msg['header']['type'] = msg_type
        if msg_flags is not None:
            msg['header']['flags'] = msg_flags

        # if there is no type & flags yet, set defaults
        # FIXME: collect usecases
        if msg['header'].get('type') is None:
            msg['header']['type'] = 0
        if msg['header'].get('flags') is None:
            msg['header']['flags'] = NLM_F_REQUEST | NLM_F_DUMP

        msg['header']['sequence_number'] = self.msg_seq
        msg['header']['pid'] = self.epid or os.getpid()
        msg.reset()
        # set fields
        if request_filter is not None:
            for field in msg.fields:
                msg[field[0]] = request_filter.get_value(
                    field[0], default=0, mode='field'
                )
            # attach NLAs
            for key, value in request_filter.items():
                nla = type(msg).name2nla(key)
                if msg.valid_nla(nla) and value is not None:
                    msg['attrs'].append([nla, value])
            # extend with custom NLAs
            if 'attrs' in request_filter:
                msg['attrs'].extend(request_filter['attrs'])
        self.msg = msg
        self.dump_filter = dump_filter
        self.terminate = terminate
        self.callback = callback
        self.command = command

    @classmethod
    def calculate_request_type(cls, command, command_map, echo=False):
        if isinstance(command, basestring):
            return (lambda x: (x[0], cls.calculate_request_flags(x[1], echo)))(
                command_map[command]
            )
        elif isinstance(command, int):
            return command, cls.calculate_request_flags('create', echo)
        elif isinstance(command, (list, tuple)):
            return command
        else:
            raise TypeError('allowed command types: int, str, list, tuple')

    @classmethod
    def calculate_request_flags(cls, mode, echo):
        return cls.flags[mode] | (
            NLM_F_ECHO if (echo and mode not in ('get', 'dump')) else 0
        )

    @staticmethod
    def match_one_message(dump_filter, msg):
        if hasattr(dump_filter, '__call__'):
            return dump_filter(msg)
        elif isinstance(dump_filter, dict):
            matches = []
            for key in dump_filter:
                # get the attribute
                if not isinstance(key, (str, tuple)):
                    continue
                value = msg.get(key)
                if value is not None and callable(dump_filter[key]):
                    matches.append(dump_filter[key](value))
                else:
                    matches.append(dump_filter[key] == value)
            return all(matches)

    async def prepare(self):
        await self.sock.setup_endpoint()
        self.msg.encode()
        self.sock.msg_queue.ensure_tag(self.msg_seq)
        if self.parser is not None:
            self.marshal.seq_map[self.msg_seq] = self.parser

    def cleanup(self):
        self.addr_pool.free(self.msg_seq, ban=0xFF)
        self.sock.msg_queue.free_tag(self.msg_seq)
        if self.msg_seq in self.marshal.seq_map:
            self.marshal.seq_map.pop(self.msg_seq)

    async def proxy(self):
        if self.sock.batch is not None:
            self.sock.batch += self.msg.data
            await self.sock.msg_queue.put(self.msg_seq, msg_done(self.msg))
            return True
        if self.sock.request_proxy is None:
            return False
        ret = self.sock.request_proxy.handle(self.msg)
        if ret == b'':
            return False
        await self.sock.msg_queue.put(self.msg_seq, ret)
        return True

    async def send(self):
        await self.prepare()
        if await self.proxy():
            return len(self.msg.data)
        count = 0
        exc = RuntimeError('Max attempts sending message')
        for count in range(30):
            try:
                return self.sock.send(self.msg.data)
            except NetlinkError as e:
                if e.code != errno.EBUSY:
                    exc = e
                    break
                log.warning(f'Error 16, retry {count}')
                await asyncio.sleep(0.3)
            except Exception as e:
                exc = e
                break
        self.cleanup()
        raise exc

    async def response(self):
        coro = self.sock.get(
            msg_seq=self.msg_seq,
            terminate=self.terminate,
            callback=self.callback,
        )
        try:
            async for msg in coro:
                if (
                    self.dump_filter is not None
                    and not self.match_one_message(self.dump_filter, msg)
                ):
                    continue
                for cr in self.sock.callbacks:
                    try:
                        if cr[0](msg):
                            cr[1](msg, *cr[2])
                    except Exception:
                        log.warning("Callback fail: %{cr}")
                yield msg
        finally:
            await coro.aclose()
            self.cleanup()


class NetlinkSocket(SyncAPI):

    async_class = AsyncNetlinkSocket

    def __init__(
        self,
        family=NETLINK_GENERIC,
        port=None,
        pid=None,
        fileno=None,
        sndbuf=1048576,
        rcvbuf=1048576,
        rcvsize=16384,
        all_ns=False,
        async_qsize=None,
        nlm_generator=True,
        target='localhost',
        ext_ack=False,
        strict_check=False,
        groups=0,
        nlm_echo=False,
        netns=None,
        flags=os.O_CREAT,
        libc=None,
        use_socket=None,
        use_event_loop=None,
        telemetry=None,
    ):
        self.asyncore = self.async_class(
            family=family,
            port=port,
            pid=pid,
            fileno=fileno,
            sndbuf=sndbuf,
            rcvbuf=rcvbuf,
            rcvsize=rcvsize,
            all_ns=all_ns,
            target=target,
            ext_ack=ext_ack,
            strict_check=strict_check,
            groups=groups,
            nlm_echo=nlm_echo,
            netns=netns,
            flags=flags,
            libc=libc,
            use_socket=use_socket,
            use_event_loop=use_event_loop,
            telemetry=telemetry,
        )
        self.asyncore.local.keep_event_loop = True
        self.asyncore.status['event_loop'] = 'new'
        self.asyncore.status['nlm_generator'] = nlm_generator
        # FIXME: temporary override from a class attribute
        if hasattr(self, 'class_gen_sync'):
            self.asyncore.status['nlm_generator'] = self.class_gen_sync
        self.asyncore.event_loop.run_until_complete(
            self.asyncore.setup_endpoint()
        )
        if self.asyncore.socket.fileno() == -1:
            raise OSError(9, 'Bad file descriptor')

    def bind(self, *argv, **kwarg):
        with self.lock:
            self.asyncore._check_tid(tag='bind', level=logging.WARN)
            self.asyncore.local.keep_event_loop = True
            self._run_with_cleanup(self.asyncore.bind, *argv, **kwarg)
            self.asyncore._register_loop_ref()

    def put(
        self,
        msg,
        msg_type,
        msg_flags=NLM_F_REQUEST,
        addr=(0, 0),
        msg_seq=None,
        msg_pid=None,
    ):
        if msg is None:
            msg_class = self.marshal.msg_map[msg_type]
            msg = msg_class()
        return self._run_with_cleanup(
            self.asyncore.put, msg, msg_type, msg_flags, addr, msg_seq, msg_pid
        )

    def nlm_request_batch(self, msgs, noraise=False):
        async def collect_data():
            return [
                x async for x in self.asyncore.nlm_request_batch(msgs, noraise)
            ]

        return self._run_with_cleanup(collect_data)

    def nlm_request(
        self,
        msg,
        msg_type,
        msg_flags=NLM_F_REQUEST | NLM_F_DUMP,
        terminate=None,
        callback=None,
        parser=None,
    ):
        ret = self._generate_with_cleanup(
            self.asyncore.nlm_request,
            msg,
            msg_type,
            msg_flags,
            terminate,
            callback,
            parser,
        )
        if self.status['nlm_generator']:
            return ret
        return tuple(ret)

    def get(self, msg_seq=0, terminate=None, callback=None, noraise=False):

        async def collect_data():
            return [
                i
                async for i in self.asyncore.get(
                    msg_seq, terminate, callback, noraise
                )
            ]

        return self._run_with_cleanup(collect_data)


class ChaoticNetlinkSocket(NetlinkSocket):
    success_rate = 1

    def __init__(self, *argv, **kwarg):
        self.success_rate = kwarg.pop('success_rate', 0.7)
        super(ChaoticNetlinkSocket, self).__init__(*argv, **kwarg)

    def get(self, *argv, **kwarg):
        if random.random() > self.success_rate:
            raise ChaoticException()
        return super(ChaoticNetlinkSocket, self).get(*argv, **kwarg)



from pyroute2 import IW, IPRoute
from pyroute2.netlink.exceptions import NetlinkError

# interface name to check
ifname = 'wlx2'
iftype = 'monitor'

iw = IW()
ip = IPRoute()
index = ip.link_lookup(ifname=ifname)[0]
try:
    print(f"Original type: '{iw.get_interface_type(index)}'")
    iw.set_interface_type(index, iftype)
    print(f"New state: '{iw.get_interface_type(index)}'")
except NetlinkError as e:
    print(f"Exception : {e}")
finally:
    iw.close()
    ip.close()



'''
IPv4 DHCP socket
================

'''

import asyncio
import logging
import socket
from typing import Optional

from pyroute2.compat import ETHERTYPE_IP
from pyroute2.dhcp.dhcp4msg import dhcp4msg
from pyroute2.dhcp.messages import ReceivedDHCPMessage, SentDHCPMessage
from pyroute2.ext.bpf import BPF
from pyroute2.ext.rawsocket import AsyncRawSocket
from pyroute2.protocols import ethmsg, ip4msg, udp4_pseudo_header, udpmsg

LOG = logging.getLogger(__name__)


UDP_HEADER_SIZE = 8
IPV4_HEADER_SIZE = 20
SKF_AD_OFF = -0x1000
SKF_AD_VLAN_TAG_PRESENT = 48


def listen_udp_port(port: int = 68) -> list[list[int]]:
    '''BPF filter that matches Ethernet + IPv4 + UDP on the given port.

    Packets tagged on a vlan are also dropped, see
    https://lore.kernel.org/netdev/51FB6A9D.2050002@redhat.com/T/
    '''
    bpf_code = [
        # Load vlan presence indicator
        [BPF.LD + BPF.B + BPF.ABS, 0, 0, SKF_AD_OFF + SKF_AD_VLAN_TAG_PRESENT],
        # bail out immediately if there is one and we don't want it
        [BPF.JMP + BPF.JEQ + BPF.K, 0, 10, 0],
        # Load eth type at offset 12 and check it's IPv4
        [BPF.LD + BPF.H + BPF.ABS, 0, 0, 12],
        [BPF.JMP + BPF.JEQ + BPF.K, 0, 8, 0x0800],
        # Load IP proto at offset 23 and check it's UDP
        [BPF.LD + BPF.B + BPF.ABS, 0, 0, 23],
        [BPF.JMP + BPF.JEQ + BPF.K, 0, 6, socket.IPPROTO_UDP],
        # load frag offset at offset 20
        [BPF.LD + BPF.H + BPF.ABS, 0, 0, 20],
        # Check mask & drop fragmented packets
        [BPF.JMP + BPF.JSET + BPF.K, 4, 0, 8191],
        # load ip header length at offset 14
        [BPF.LDX + BPF.B + BPF.MSH, 0, 0, 14],
        # load udp dport from that offset + 16 and check it
        [BPF.LD + BPF.H + BPF.IND, 0, 0, 16],
        [BPF.JMP + BPF.JEQ + BPF.K, 0, 1, port],
        # allow packet
        [BPF.RET + BPF.K, 0, 0, 65535],
        # drop packet
        [BPF.RET + BPF.K, 0, 0, 0],
    ]
    return bpf_code


class AsyncDHCP4Socket(AsyncRawSocket):
    '''
    Parameters:

    * ifname -- interface name to work on
    * port -- UDP port to listen on

    This raw socket binds to an interface and installs a BPF filter
    to receive (non-VLAN) messages only on the specified UDP port.
    It can be used in poll/select and implements the async context manager
    protocol, so can be used in `aysnc with` statements.
    '''

    def __init__(self, ifname: str, port: int = 68):
        AsyncRawSocket.__init__(self, ifname, listen_udp_port(port))
        self.port = port
        self._loop = asyncio.get_running_loop()

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        # We define this as a property because it's easier to patch in tests
        return self._loop

    async def put(self, msg: SentDHCPMessage) -> SentDHCPMessage:
        '''Send a DHCP message.

        This encapsulates the `SentDHCPMessage` into Ethernet, IPv4, and UDP.

        If not provided, both the Ethernet source and the DHCP chaddr are set
        to the MAC address of the underlying socket's interface.

        Example::

            msg = SentDHCPMessage(
                dhcp=dhcp4msg({
                    'op': bootp.MessageType.BOOTREQUEST,
                    'flags': bootp.Flag.BROADCAST,
                    'options': {
                        'message_type': dhcp.MessageType.DISCOVER,
                        'parameter_list': [
                            dhcp.Option.SUBNET_MASK,
                            dhcp.Option.ROUTER,
                            dhcp.Option.NAME_SERVER,
                            dhcp.Option.LEASE_TIME,
                        ],
                    },
                }),
            )
            await sock.put(msg)

        '''

        if msg.sport != self.port:
            raise ValueError(
                f"Client source port is set to {self.port}, "
                f"cannot send message from port {msg.sport}."
            )

        if not msg.eth_src:
            msg.eth_src = self.l2addr

        # DHCP layer
        dhcp = msg.dhcp

        # auto add src addr
        if dhcp['chaddr'] is None:
            dhcp['chaddr'] = msg.eth_src

        data = dhcp.encode().buf
        dhcp_payload_size = len(data)

        # UDP layer
        udp = udpmsg(
            {
                'sport': self.port,
                'dport': msg.dport,
                'len': UDP_HEADER_SIZE + dhcp_payload_size,
            }
        )
        # Pseudo UDP header, only for checksum purposes
        udph = udp4_pseudo_header(
            {
                'src': msg.ip_src,
                'dst': msg.ip_dst,
                'len': UDP_HEADER_SIZE + dhcp_payload_size,
            }
        )
        udp['csum'] = self.csum(udph.encode().buf + udp.encode().buf + data)
        udp.reset()

        # IPv4 layer
        ip4 = ip4msg(
            {
                'len': IPV4_HEADER_SIZE + UDP_HEADER_SIZE + dhcp_payload_size,
                'proto': socket.IPPROTO_UDP,
                'dst': msg.ip_dst,
                'src': msg.ip_src,
            }
        )
        ip4['csum'] = self.csum(ip4.encode().buf)
        ip4.reset()

        # MAC layer
        eth = ethmsg(
            {'dst': msg.eth_dst, 'src': msg.eth_src, 'type': ETHERTYPE_IP}
        )

        data = eth.encode().buf + ip4.encode().buf + udp.encode().buf + data
        await self.loop.sock_sendall(self, data)
        dhcp.reset()
        return msg

    async def get(self) -> ReceivedDHCPMessage:
        '''
        Get the next incoming packet from the socket and try
        to decode it as IPv4 DHCP.

        Packets that cannot be decoded are logged & discarded.
        Invalid/truncated packets raise `ValueError`.

        Example::

            # Send a DISCOVER and read an OFFER
            disco = messages.discover(parameter_list=[
                dhcp.Option.SUBNET_MASK,
                dhcp.Option.ROUTER,
                dhcp.Option.NAME_SERVER,
                dhcp.Option.LEASE_TIME,
            ])
            await sock.put(disco)
            offer = await sock.get()
            print('received', offer.message_type,
                  'from', offer.eth_src,
                  'lease time', offer.dhcp['options']['lease_time'],
            )

        '''
        msg: Optional[ReceivedDHCPMessage] = None
        while not msg:
            raw = await self.loop.sock_recv(self, 4096)
            try:
                msg = self._decode_msg(raw)
            except ValueError as err:
                LOG.error('%s', err)
        return msg

    @classmethod
    def _decode_msg(cls, data: bytes) -> ReceivedDHCPMessage:
        '''Decode an IPv4 DHCP packet from bytes.

        No analysis is done here. The MAC/IPv4/UDP headers are stripped out,
        the relevant values are stored in the `ReceivedDHCPMessage` metadata,
        and the rest is interpreted as DHCP.
        '''
        eth = ethmsg(buf=data).decode()
        ip4 = ip4msg(buf=data, offset=eth.offset).decode()
        udp = udpmsg(buf=data, offset=ip4.offset).decode()
        dhcp = dhcp4msg(buf=data, offset=udp.offset).decode()
        return ReceivedDHCPMessage(
            dhcp=dhcp,
            eth_src=eth['src'],
            eth_dst=eth['dst'],
            ip_src=ip4['src'],
            ip_dst=ip4['dst'],
            sport=udp['sport'],
            dport=udp['dport'],
        )



'''
IPDB
----

Deprecated since 0.7.12, removed from the library.

The current `pyroute2.ipdb` implementation is a wrapper around NDB
only to provide minimal compatibility for the old legacy code. Do
NOT use IPDB.
'''

import errno
import logging
import warnings

from pyroute2.ndb.main import NDB
from pyroute2.netlink.exceptions import NetlinkError

log = logging.getLogger(__name__)


class CreateException(Exception):
    pass


class CommitException(Exception):
    pass


class ObjectProxy(dict):

    _translate_keys = {}

    def __init__(self, obj, ready=True):
        self._obj = obj
        self._ready = ready

    def __getattribute__(self, key):
        if key[:4] == 'set_':

            def set_value(value):
                self[key[4:]] = value
                return self

            return set_value
        try:
            return self[key]
        except KeyError:
            return super(ObjectProxy, self).__getattribute__(key)

    def __setattr__(self, key, value):
        if key in ('_obj', '_ready', '_translate_keys'):
            super(ObjectProxy, self).__setattr__(key, value)
        else:
            super(ObjectProxy, self).__getattribute__('_obj')[key] = value

    def __getitem__(self, key):
        tk = super().__getattribute__('_translate_keys')
        if isinstance(key, str) and key in tk:
            return super().__getattribute__('_obj')[tk[key](self)]
        return super(ObjectProxy, self).__getattribute__('_obj')[key]

    def __setitem__(self, key, value):
        tk = super().__getattribute__('_translate_keys')
        if isinstance(key, str) and key in tk:
            super().__getattribute__('_obj')[tk[key](self)] = value
        super(ObjectProxy, self).__getattribute__('_obj')[key] = value

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if hasattr(self, 'commit'):
            self.commit()

    def __repr__(self):
        return repr(super(ObjectProxy, self).__getattribute__('_obj'))

    def __contains__(self, key):
        return key in super(ObjectProxy, self).__getattribute__('_obj')

    def get_ndb_object(self):
        return self._obj

    def get(self, key, *argv):
        return self._obj.get(key, *argv)

    def keys(self):
        return self._obj.keys()

    def items(self):
        return self._obj.items()

    def values(self):
        return self._obj.values()

    def __iter__(self):
        return self._obj.__iter__()

    @property
    def _mode(self):
        return 'implicit'


class Interface(ObjectProxy):

    _translate_keys = {'mode': lambda x: f'{x["kind"]}_mode'}

    def add_ip(self, address=None, prefixlen=None, **kwarg):
        if address is not None:
            kwarg['address'] = address
        if prefixlen is not None:
            kwarg['prefixlen'] = prefixlen
        self._obj.add_ip(spec=kwarg)
        return self

    def del_ip(self, address=None, prefixlen=None, **kwarg):
        if address is not None:
            kwarg['address'] = address
        if prefixlen is not None:
            kwarg['prefixlen'] = prefixlen
        self._obj.del_ip(spec=kwarg)
        return self

    def add_port(self, *argv, **kwarg):
        self._obj.add_port(*argv, **kwarg)
        return self

    def del_port(self, *argv, **kwarg):
        self._obj.del_port(*argv, **kwarg)
        return self

    def commit(self, *argv, **kwarg):
        try:
            self._obj.commit(*argv, **kwarg)
        except Exception as e:
            if self._ready:
                raise CommitException(e)
            else:
                raise CreateException(e)
        self._ready = True
        return self

    def up(self):
        self._obj.set('state', 'up')
        return self

    def down(self):
        self._obj.set('state', 'down')
        return self

    def remove(self):
        self._obj.remove()
        return self

    @property
    def if_master(self):
        return self._obj.get('master', None)

    @property
    def ipaddr(self):
        report = self._obj.ipaddr.dump()
        report.select_fields('address', 'prefixlen')
        return tuple(report)


class Interfaces(ObjectProxy):
    text_create = '''
When `create().commit()` fails, the failed interface object behaves
differently in IPDB and NDB. IPDB saves the failed object in the database,
while the NDB database contains only the system reflection, and the failed
object may stay only being referenced by a variable.
'''

    def __getitem__(self, key):
        return Interface(super(Interfaces, self).__getitem__(key))

    def __iter__(self):
        return iter(self.keys())

    def add(self, *argv, **kwarg):
        return self.create(*argv, **kwarg)

    def get(self, spec, *argv):
        try:
            return self[spec]
        except KeyError:
            if len[argv] > 0:
                return argv[0]
            raise

    def create(self, *argv, **kwarg):
        log.warning(self.text_create)
        key = dict(
            filter(lambda x: x[0] in ('ifname', 'index'), kwarg.items())
        )
        if key in self:
            if kwarg.get('reuse'):
                return self[key]
            raise CreateException(NetlinkError(errno.EEXIST, 'object exists'))
        return Interface(self._obj.create(*argv, **kwarg), ready=False)

    def keys(self):
        ret = []
        for record in self._obj.dump():
            ret += [record.ifname, record.index]
        return ret

    def has_key(self, key):
        return key in self.keys()


class IPDB(object):
    text_create = '''
IPDB has a shortcut method to create interfaces: `ipdb.create(...)`.

NDB has `create()` methods only under respective views:
`ndb.interfaces.create(...)`, `ndb.addresses.create(...)` etc.
'''

    text_nl = '''
Unlike IPDB, NDB can work with many netlink sources. The default one
referenced as `localhost`::

    #
    # these two statements are equivalent:
    #
    ndb.sources['localhost'].nl.get_links()
    ipdb.nl.get_links()

'''

    def __init__(self, *argv, **kwarg):
        warnings.warn(
            '''
            IPDB module is deprecated and removed.
            This IPDB instance is just a wrapper around
            NDB. See more:

            https://github.com/svinota/pyroute2/wiki/IPDB-EOL
            ''',
            DeprecationWarning,
        )
        sources = kwarg.pop('sources', [{'target': 'localhost'}])
        if argv or kwarg:
            log.warning(
                '%s does not support IPDB parameters, ignoring',
                self.__class__.__name__,
            )
        if len(argv) > 0 or 'nl' in kwarg:
            log.warning(
                '%s does not support shared netlink sources,'
                ' ignoring `nl` and starting with local IPRoute',
                self.__class__.__name__,
            )

        self._ndb = NDB(sources=sources)
        self.interfaces = Interfaces(self._ndb.interfaces)

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.release()

    @property
    def nl(self):
        log.warning(self.text_nl)
        return self._ndb.sources['localhost'].nl

    @property
    def ipaddr(self):
        ret = dict([(x.index, []) for x in self._ndb.interfaces.dump()])
        for record in self._ndb.addresses.dump():
            ret[record.index].append((record.address, record.prefixlen))
        return ret

    def create(self, *argv, **kwarg):
        log.warning(self.text_create)
        return self.interfaces.create(*argv, **kwarg)

    def release(self):
        self._ndb.close()



import pprint
import sys

from pyroute2.netlink.generic.ethtool import NlEthtool

if len(sys.argv) != 2:
    raise Exception("USAGE: {0} IFNAME".format(sys.argv[0]))


IFNAME = sys.argv[1]
eth = NlEthtool()

print("kernel ok?:", eth.is_nlethtool_in_kernel())
pprint.pprint(eth.get_linkmode(IFNAME))
print("")
pprint.pprint(eth.get_linkinfo(IFNAME))
print("")
pprint.pprint(eth.get_stringset(IFNAME))
print("")
pprint.pprint(eth.get_linkstate(IFNAME))
print("")
pprint.pprint(eth.get_wol(IFNAME))



import pytest

from pyroute2.netlink import NlaMapAdapter, NlaSpec, nlmsg_atoms, rtnl
from pyroute2.netlink.rtnl.ifaddrmsg import ifaddrmsg
from pyroute2.netlink.rtnl.marshal import MarshalRtnl

sample_data_ipaddr = (
    b'L\x00\x00\x00\x14\x00\x02\x00\xff\x00\x00\x00\xfb\x8d\x00\x00\x02\x08'
    b'\x80\xfe\x01\x00\x00\x00\x08\x00\x01\x00\x7f\x00\x00\x01\x08\x00\x02\x00'
    b'\x7f\x00\x00\x01\x07\x00\x03\x00lo\x00\x00\x08\x00\x08\x00\x80\x00\x00'
    b'\x00\x14\x00\x06\x00\xff\xff\xff\xff\xff\xff\xff\xffQ\x00\x00\x00Q\x00'
    b'\x00\x00'
)


class ifaddrmsg_default_decode(ifaddrmsg):
    # same function will be used both for decode and encode
    nla_map = NlaMapAdapter(
        lambda x: NlaSpec(nlmsg_atoms.hex, x, f'IFA_NLA_{x}')
    )


class ifaddrmsg_dict_decode(ifaddrmsg):
    # define separate decode / encode adapters
    nla_map = {
        'decode': NlaMapAdapter(
            lambda x: NlaSpec(nlmsg_atoms.hex, x, f'IFA_NLA_{x}')
        ),
        'encode': None,
    }


@pytest.mark.parametrize(
    'nlmsg_class,data',
    (
        (ifaddrmsg_default_decode, sample_data_ipaddr),
        (ifaddrmsg_dict_decode, sample_data_ipaddr),
    ),
    ids=['default_decode', 'dict_decode'],
)
def test_decode_adapter(nlmsg_class, data):
    marshal = MarshalRtnl()
    marshal.msg_map[rtnl.RTM_NEWADDR] = nlmsg_class
    msgs = tuple(marshal.parse(data))
    msg = msgs[0]
    assert len(msgs) == 1
    assert msg.get_attr('IFA_NLA_1') == '7f:00:00:01'  # IFA_ADDRESS
    assert msg.get_attr('IFA_NLA_2') == '7f:00:00:01'  # IFA_LOCAL
    assert msg.get_attr('IFA_NLA_3') == '6c:6f:00'  # IFA_LABEL
    assert (
        msg.get_attr('IFA_NLA_6')  # IFA_CACHEINFO
        == 'ff:ff:ff:ff:ff:ff:ff:ff:51:00:00:00:51:00:00:00'
    )
    assert msg.get_attr('IFA_NLA_8') == '80:00:00:00'  # IFA_FLAGS



from pyroute2.common import map_namespace

# ARP protocol HARDWARE identifiers.
ARPHRD_NETROM = 0  # from KA9Q: NET/ROM pseudo
ARPHRD_ETHER = 1  # Ethernet 10Mbps
ARPHRD_EETHER = 2  # Experimental Ethernet
ARPHRD_AX25 = 3  # AX.25 Level 2
ARPHRD_PRONET = 4  # PROnet token ring
ARPHRD_CHAOS = 5  # Chaosnet
ARPHRD_IEEE802 = 6  # IEEE 802.2 Ethernet/TR/TB
ARPHRD_ARCNET = 7  # ARCnet
ARPHRD_APPLETLK = 8  # APPLEtalk
ARPHRD_DLCI = 15  # Frame Relay DLCI
ARPHRD_ATM = 19  # ATM
ARPHRD_METRICOM = 23  # Metricom STRIP (new IANA id)
ARPHRD_IEEE1394 = 24  # IEEE 1394 IPv4 - RFC 2734
ARPHRD_EUI64 = 27  # EUI-64
ARPHRD_INFINIBAND = 32  # InfiniBand

# Dummy types for non ARP hardware
ARPHRD_SLIP = 256
ARPHRD_CSLIP = 257
ARPHRD_SLIP6 = 258
ARPHRD_CSLIP6 = 259
ARPHRD_RSRVD = 260  # Notional KISS type
ARPHRD_ADAPT = 264
ARPHRD_ROSE = 270
ARPHRD_X25 = 271  # CCITT X.25
ARPHRD_HWX25 = 272  # Boards with X.25 in firmware
ARPHRD_PPP = 512
ARPHRD_CISCO = 513  # Cisco HDLC
ARPHRD_HDLC = ARPHRD_CISCO
ARPHRD_LAPB = 516  # LAPB
ARPHRD_DDCMP = 517  # Digital's DDCMP protocol
ARPHRD_RAWHDLC = 518  # Raw HDLC

ARPHRD_TUNNEL = 768  # IPIP tunnel
ARPHRD_TUNNEL6 = 769  # IP6IP6 tunnel
ARPHRD_FRAD = 770  # Frame Relay Access Device
ARPHRD_SKIP = 771  # SKIP vif
ARPHRD_LOOPBACK = 772  # Loopback device
ARPHRD_LOCALTLK = 773  # Localtalk device
ARPHRD_FDDI = 774  # Fiber Distributed Data Interface
ARPHRD_BIF = 775  # AP1000 BIF
ARPHRD_SIT = 776  # sit0 device - IPv6-in-IPv4
ARPHRD_IPDDP = 777  # IP over DDP tunneller
ARPHRD_IPGRE = 778  # GRE over IP
ARPHRD_PIMREG = 779  # PIMSM register interface
ARPHRD_HIPPI = 780  # High Performance Parallel Interface
ARPHRD_ASH = 781  # Nexus 64Mbps Ash
ARPHRD_ECONET = 782  # Acorn Econet
ARPHRD_IRDA = 783  # Linux-IrDA
# ARP works differently on different FC media .. so
ARPHRD_FCPP = 784  # Point to point fibrechannel
ARPHRD_FCAL = 785  # Fibrechannel arbitrated loop
ARPHRD_FCPL = 786  # Fibrechannel public loop
ARPHRD_FCFABRIC = 787  # Fibrechannel fabric
# 787->799 reserved for fibrechannel media types
ARPHRD_IEEE802_TR = 800  # Magic type ident for TR
ARPHRD_IEEE80211 = 801  # IEEE 802.11
ARPHRD_IEEE80211_PRISM = 802  # IEEE 802.11 + Prism2 header
ARPHRD_IEEE80211_RADIOTAP = 803  # IEEE 802.11 + radiotap header
ARPHRD_MPLS_TUNNEL = 899  # MPLS Tunnel Interface

ARPHRD_VOID = 0xFFFF  # Void type, nothing is known
ARPHRD_NONE = 0xFFFE  # zero header length

(ARPHRD_NAMES, ARPHRD_VALUES) = map_namespace("ARPHRD_", globals())



from collections import OrderedDict

from pyroute2.common import get_address_family
from pyroute2.netlink.rtnl.fibmsg import fibmsg

from ..objects import AsyncObject

schema = fibmsg.sql_schema().unique_index(
    'family',
    'dst_len',
    'src_len',
    'tos',
    'action',
    'flags',
    'FRA_DST',
    'FRA_SRC',
    'FRA_IIFNAME',
    'FRA_GOTO',
    'FRA_PRIORITY',
    'FRA_FWMARK',
    'FRA_FLOW',
    'FRA_TUN_ID',
    'FRA_SUPPRESS_IFGROUP',
    'FRA_SUPPRESS_PREFIXLEN',
    'FRA_TABLE',
    'FRA_FWMASK',
    'FRA_OIFNAME',
    'FRA_L3MDEV',
    'FRA_UID_RANGE',
    'FRA_PROTOCOL',
    'FRA_IP_PROTO',
    'FRA_SPORT_RANGE',
    'FRA_DPORT_RANGE',
)

init = {
    'specs': [['rules', schema]],
    'classes': [['rules', fibmsg]],
    'event_map': {fibmsg: ['rules']},
}


class Rule(AsyncObject):
    table = 'rules'
    msg_class = fibmsg
    api = 'rule'
    _replace_on_key_change = True

    @classmethod
    def summary(cls, view):
        req = '''
              SELECT
                f_target, f_tflags, f_family,
                f_FRA_PRIORITY, f_action, f_FRA_TABLE
              FROM
                rules
              '''
        yield ('target', 'tflags', 'family', 'priority', 'action', 'table')
        for record in view.ndb.schema.fetch(req):
            yield record

    def __init__(self, *argv, **kwarg):
        kwarg['iclass'] = fibmsg
        self._fields = [x[0] for x in fibmsg.fields]
        self.event_map = {fibmsg: "load_rtnlmsg"}
        key = argv[1]
        families = set(
            [
                get_address_family(x)
                for x in (key.get('src'), key.get('dst'))
                if x != '' and x is not None
            ]
        )
        if len(families) > 1:
            raise TypeError('src and dst must be of the same IP family')
        elif len(families) == 1:
            key['family'] = families.pop()
        super(Rule, self).__init__(*argv, **kwarg)

    def load_sql(self, *argv, **kwarg):
        spec = super(Rule, self).load_sql(*argv, **kwarg)
        if spec is None:
            return
        nkey = OrderedDict()
        for name_norm, name_raw, value in zip(self.names, self.spec, spec):
            if name_raw in self.kspec:
                nkey[name_raw] = value
            if name_norm not in self._fields and value in (0, ''):
                dict.__setitem__(self, name_norm, None)
        self._key = nkey
        return spec



import errno
import struct
import threading
from functools import partial
from typing import Callable, Generator, Optional

from pyroute2.netlink import (
    NLM_F_ACK_TLVS,
    NLMSG_DONE,
    NLMSG_ERROR,
    mtypes,
    nlmsg,
    nlmsgerr,
)
from pyroute2.netlink.exceptions import (
    NetlinkDecodeError,
    NetlinkError,
    NetlinkHeaderDecodeError,
)


class Marshal:
    '''
    Generic marshalling class
    '''

    msg_map = {}
    seq_map = None
    key_offset = None
    key_format = None
    key_mask = None
    debug = False
    default_message_class = nlmsg
    error_type = NLMSG_ERROR

    def __init__(self):
        self.lock = threading.Lock()
        self.msg_map = self.msg_map.copy()
        self.seq_map = {}
        self.defragmentation = {}

    def parse_one_message(
        self, key, flags, sequence_number, data, offset, length
    ):
        msg = None
        error = None
        msg_class = self.msg_map.get(key, self.default_message_class)
        # ignore length for a while
        # get the message
        if (key == self.error_type) or (
            key == NLMSG_DONE and flags & NLM_F_ACK_TLVS
        ):
            msg = nlmsgerr(data, offset=offset)
        else:
            msg = msg_class(data, offset=offset)

        try:
            msg.decode()
        except NetlinkHeaderDecodeError as e:
            msg = nlmsg()
            msg['header']['error'] = e
        except NetlinkDecodeError as e:
            msg['header']['error'] = e

        if isinstance(msg, nlmsgerr) and msg['error'] != 0:
            code = abs(msg['error'])
            if code == errno.ENOBUFS:
                error = OSError(code, msg.get_attr('NLMSGERR_ATTR_MSG'))
            else:
                error = NetlinkError(code, msg.get_attr('NLMSGERR_ATTR_MSG'))
                enc_type = struct.unpack_from('H', data, offset + 24)[0]
                enc_class = self.msg_map.get(enc_type, nlmsg)
                enc = enc_class(data, offset=offset + 20)
                enc.decode()
                msg['header']['errmsg'] = enc

        msg['header']['error'] = error
        return msg

    def get_parser(self, key, flags, sequence_number):
        return self.seq_map.get(
            sequence_number,
            partial(self.parse_one_message, key, flags, sequence_number),
        )

    def parse(
        self,
        data: bytes,
        seq: Optional[int] = None,
        callback: Optional[Callable] = None,
        skip_alien_seq: bool = False,
    ) -> Generator[nlmsg, None, None]:
        '''
        Parse string data.

        At this moment all transport, except of the native
        Netlink is deprecated in this library, so we should
        not support any defragmentation on that level
        '''
        offset = 0
        # there must be at least one header in the buffer,
        # 'IHHII' == 16 bytes
        while offset <= len(data) - 16:
            # pick type and length
            (length, key, flags, sequence_number) = struct.unpack_from(
                'IHHI', data, offset
            )
            if skip_alien_seq and sequence_number != seq:
                continue
            if not 0 < length <= len(data):
                break
            # support custom parser keys
            # see also: pyroute2.netlink.diag.MarshalDiag
            if self.key_format is not None:
                (key,) = struct.unpack_from(
                    self.key_format, data, offset + self.key_offset
                )
                if self.key_mask is not None:
                    key &= self.key_mask

            parser = self.get_parser(key, flags, sequence_number)
            msg = parser(data, offset, length)
            offset += length
            if msg is None:
                continue

            if callable(callback) and seq == sequence_number:
                try:
                    if callback(msg):
                        continue
                except Exception:
                    pass

            mtype = msg['header'].get('type', None)
            if mtype in (1, 2, 3, 4) and 'event' not in msg:
                msg['event'] = mtypes.get(mtype, 'none')
            self.fix_message(msg)
            yield msg

    def is_enough(self, msg):
        return msg['header']['type'] == NLMSG_DONE

    def fix_message(self, msg):
        pass



import asyncio
import json

from pyroute2.netlink.core import AsyncCoreSocket, CoreMessageQueue
from pyroute2.netlink.coredata import CoreConfig, CoreSocketSpec
from pyroute2.plan9 import (
    Marshal9P,
    Plan9Exit,
    Stat,
    Tattach,
    Tauth,
    Tcall,
    Tclunk,
    Tcreate,
    Topen,
    Tread,
    Tremove,
    Tstat,
    Tversion,
    Twalk,
    Twrite,
    Twstat,
    msg_rattach,
    msg_rcall,
    msg_rclunk,
    msg_rerror,
    msg_ropen,
    msg_rread,
    msg_rstat,
    msg_rversion,
    msg_rwalk,
    msg_rwrite,
    msg_rwstat,
)
from pyroute2.plan9.filesystem import Filesystem, Session

data = str(dir())


def get_exception_args(exc):
    args = []
    if hasattr(exc, 'errno'):
        args.append(exc.errno)
        args.append(exc.strerror)
    return args


def route(rtable, request, state):
    def decorator(f):
        rtable[request] = f
        return f

    return decorator


class Plan9ServerProtocol(asyncio.Protocol):
    rtable = {}

    def __init__(self, on_con_lost, marshal, filesystem):
        self.transport = None
        self.session = None
        self.filesystem = filesystem
        self.marshal = marshal
        self.on_con_lost = on_con_lost

    @route(rtable, request=Tversion, state=(None,))
    def t_version(self, req):
        m = msg_rversion()
        m['header']['tag'] = 0xFFFF
        m['msize'] = req['msize']
        m['version'] = '9P2000'
        return m

    @route(rtable, request=Tauth, state=(Tversion,))
    def t_auth(self, req):
        m = msg_rerror()
        m['ename'] = 'no authentication required'
        return m

    @route(rtable, request=Tattach, state=(Tauth,))
    def t_attach(self, req):
        m = msg_rattach()
        root = self.session.filesystem.inodes[0]
        self.session.set_fid(req['fid'], root)
        m['qid'] = root.qid
        return m

    @route(rtable, request=Twalk, state=(Tattach,))
    def t_walk(self, req):
        m = msg_rwalk()
        inode = self.session.get_fid(req['fid'])
        wqid = []
        if len(req['wname']) == 0:
            self.session.set_fid(req['newfid'], inode)
        else:
            for name in req['wname']:
                if name == '.':
                    continue
                elif name == '..':
                    inode = inode.get_parent()
                else:
                    inode = inode.get_child(name)
                wqid.append(inode.qid)
        m['wqid'] = wqid
        self.session.set_fid(req['newfid'], inode)
        return m

    @route(rtable, request=Tstat, state=(Twalk,))
    def t_stat(self, req):
        m = msg_rstat()
        inode = self.session.get_fid(req['fid'])
        inode.sync()
        m['stat'] = inode.stat
        return m

    @route(rtable, request=Twstat, state=(Twalk,))
    def t_wstat(self, req):
        m = msg_rwstat()
        return m

    @route(rtable, request=Topen, state=(Twalk, Tstat))
    def t_open(self, req):
        m = msg_ropen()
        m['qid'] = self.session.get_fid(req['fid']).qid
        m['iounit'] = 8192
        return m

    @route(rtable, request=Tcall, state=(Twalk, Topen, Tstat))
    def t_call(self, req):
        m = msg_rcall()
        inode = self.session.get_fid(req['fid'])
        m['err'] = 255
        if Tcall in inode.callbacks:
            m = inode.callbacks[Tcall](self.session, inode, req, m)
        return m

    @route(rtable, request=Twrite, state=(Topen,))
    def t_write(self, req):
        m = msg_rwrite()
        inode = self.session.get_fid(req['fid'])
        if Twrite in inode.callbacks:
            return inode.callbacks[Twrite](self.session, inode, req, m)
        if inode.qid['type'] & 0x80:
            raise TypeError('can not call write() on dir')
        inode.data.seek(req['offset'])
        m['count'] = inode.data.write(req['data'])
        return m

    @route(rtable, request=Tread, state=(Topen,))
    def t_read(self, req):
        m = msg_rread()
        inode = self.session.get_fid(req['fid'])
        if Tread in inode.callbacks:
            return inode.callbacks[Tread](self.session, inode, req, m)
        if inode.qid['type'] & 0x80:
            data = bytearray()
            offset = 0
            for child in inode.children:
                offset = Stat.encode_into(data, offset, child.stat)
            data = data[req['offset'] : req['offset'] + req['count']]
        else:
            inode.data.seek(req['offset'])
            data = inode.data.read(req['count'])
        m['data'] = data
        return m

    @route(rtable, request=Tclunk, state=(Topen, Tstat, Twalk, Tread))
    def t_clunk(self, req):
        return msg_rclunk()

    @route(rtable, request=Tcreate, state=(Twalk,))
    def t_create(self, req):
        return self.permission_denied(req)

    @route(rtable, request=Tremove, state=(Twalk,))
    def t_remove(self, req):
        return self.permission_denied(req)

    def permission_denied(self, req):
        r_message = msg_rerror()
        r_message['ename'] = 'permission denied'
        r_message['header']['tag'] = req['header']['tag']
        return r_message

    def error(self, e, tag=0):
        r_message = msg_rerror()
        spec = {
            'class': e.__class__.__name__,
            'argv': get_exception_args(e),
            'str': str(e),
        }
        r_message['ename'] = json.dumps(spec)
        r_message['header']['tag'] = tag
        r_message.encode()
        self.transport.write(r_message.data)

    def data_received(self, data):
        for t_message in self.marshal.parse(data):
            tag = t_message['header']['tag']
            try:
                r_message = self.rtable[t_message['header']['type']](
                    self, t_message
                )
                r_message['header']['tag'] = tag
                r_message.encode()
            except Plan9Exit as e:
                self.error(e, tag)
                self.transport.abort()
                self.transport.close()
                return
            except Exception as e:
                self.error(e, tag)
                return
            self.transport.write(r_message.data)

    def connection_made(self, transport):
        self.transport = transport
        self.session = Session(self.filesystem)


class Plan9ServerSocket(AsyncCoreSocket):
    '''9p2000 server.

    Requires either an IP address to listen on, or an open
    `SOCK_STREAM` socket to operate. An IP example, suitable
    to establish IPC between processes in one network:

    .. testcode::

        from pyroute2 import Plan9ClientSocket, Plan9ServerSocket

        address = ('localhost', 8149)
        p9server = Plan9ServerSocket(address=address)
        p9client = Plan9ClientSocket(address=address)

    Server/client running on a `socketpair()` suitable
    for internal API within one process, or between
    parent/child processes:

    .. testcode::

        from socket import socketpair

        from pyroute2 import Plan9ClientSocket, Plan9ServerSocket

        server, client = socketpair()
        p9server = Plan9ServerSocket(use_socket=server)
        p9client = Plan9ClientSocket(use_socket=client)


    '''

    def __init__(self, address=None, use_socket=None, use_event_loop=None):
        self.spec = CoreSocketSpec(
            CoreConfig(
                tag_field='tag',
                target='localhost',
                netns=None,
                address=address,
                use_socket=use_socket is not None,
                use_event_loop=use_event_loop is not None,
            )
        )
        self.filesystem = Filesystem()
        self.marshal = Marshal9P()
        super().__init__(use_socket=use_socket, use_event_loop=use_event_loop)

    def register_function(
        self,
        func,
        inode,
        loader=json.loads,
        dumper=lambda x: json.dumps(x).encode('utf-8'),
    ):
        '''Register a function to an file.

        The file usage:

        * `write()`: write arguments for the call as a json dictionary of
          keyword arguments to the file data.
        * `read()`:
            1. if the arguments were written to the data, call the function
               and write the result to the file data
            2. read the file data and return to the client
        * `call()`: protocol extension, `Tcall` = 80, `Rcall` = 81, make
          this in one turn.

        .. testcode::
            :hide:

            from pyroute2.plan9 import Tcall, Rcall

            assert Tcall == 80
            assert Rcall == 81

        Registering a function:

        .. testcode::

            def communicate(a, b):
                return a + b


            def example_register():
                fd = p9server.filesystem.create('test_func')
                p9server.register_function(communicate, fd)

        Communication using Twrite/Tread:

        .. testcode::

            import json


            async def example_write():
                fid = await p9client.fid('test_func')
                await p9client.write(
                    fid,
                    json.dumps({"a": 17, "b": 25})
                )
                msg = await p9client.read(fid)
                response = json.loads(msg['data'])
                assert response == 42

        Same, using a command line 9p client from plan9port::

            $ echo '{"a": 17, "b": 25}' | 9p -a localhost:8149 write test_func
            $ 9p -a localhost:8149 read test_func
            42

        And using a mounted file system via FUSE client from plan9port::

            $ 9pfuse localhost:8149 mnt
            $ echo '{"a": 17, "b": 25}' >mnt/test_func
            $ cat mnt/test_func
            42

        And the same, but using Tcall:

        .. testcode::

            async def example_call():
                fid = await p9client.fid('test_func')
                response = await p9client.call(fid, argv=(17, 25))
                assert response == 42

        And finnaly run this code:

        .. testcode::

            async def main():
                server_task = await p9server.async_run()
                example_register()
                await p9client.start_session()
                await example_write()
                await example_call()
                server_task.cancel()

            asyncio.run(main())
        '''
        return inode.register_function(func, loader, dumper)

    async def setup_endpoint(self):
        if getattr(self.local, 'server', None) is not None:
            return
        self.local.msg_queue = CoreMessageQueue(event_loop=self.event_loop)
        if self.status['use_socket']:
            self.local.server = None
            self.local.transport, self.local.protocol = (
                await self.event_loop.create_connection(
                    lambda: Plan9ServerProtocol(
                        self.connection_lost, self.marshal, self.filesystem
                    ),
                    sock=self.use_socket,
                )
            )
        else:
            self.local.transport = None
            self.local.protocol = None
            self.local.server = await self.event_loop.create_server(
                lambda: Plan9ServerProtocol(
                    self.connection_lost, self.marshal, self.filesystem
                ),
                *self.status['address'],
            )

    async def async_run(self):
        '''Return the server asyncio task.

        Using this task one can stop the server:

        .. testcode::

            async def main():
                server = Plan9ServerSocket(address=('localhost', 8149))
                server_task = await server.async_run()
                # ... server is running here
                server_task.cancel()
                # ... server is stopped

            asyncio.run(main())

        To forcefully close all client connections and stop the server
        immediately from a registered function, one can pass this task
        to the function, cancel it, and raise `Plan9Exit()` exception:

        .. testcode::

            import functools

            from pyroute2.plan9 import Plan9Exit

            server_sock, client_sock = socketpair()


            def test_exit_func(context):
                if 'server_task' in context:
                    context['server_task'].cancel()
                    raise Plan9Exit('server stopped upon client request')
                return 'server starting, please wait'


            async def server():
                p9server = Plan9ServerSocket(use_socket=server_sock)
                context = {}

                inode = p9server.filesystem.create('stop')
                p9server.register_function(
                    functools.partial(test_exit_func, context),
                    inode
                )
                context['server_task'] = await p9server.async_run()

                try:
                    await context['server_task']
                except asyncio.exceptions.CancelledError:
                    pass

                assert context['server_task'].cancelled()

        .. testcode::
            :hide:

            async def client():
                p9client = Plan9ClientSocket(use_socket=client_sock)
                await p9client.start_session()
                fid = await p9client.fid('stop')
                try:
                    await p9client.call(fid)
                except Plan9Exit:
                    pass


            async def main():
                await asyncio.gather(server(), client())

            asyncio.run(main())
        '''
        await self.setup_endpoint()
        if self.status['use_socket']:
            return self.protocol.on_con_lost
        else:
            return asyncio.create_task(self.local.server.serve_forever())

    def run(self):
        '''A simple synchronous runner.

        Uses `event_loop.run_forever()`.
        '''
        self.event_loop.create_task(self.async_run())
        self.event_loop.run_forever()



from pyroute2 import NDB
from pyroute2.common import uifname

with NDB() as ndb:

    # dummy, bridge and bond interfaces are created in the
    # same way
    #
    # uifname() function is used here only to generate a
    # unique name of the interface for the regression testing,
    # you can pick up any name
    #
    ifname = uifname()
    (
        ndb.interfaces.create(kind='dummy', ifname=ifname, state='up')
        .set('state', 'up')
        .set('address', '00:11:22:33:44:55')
        .commit()
    )
    print(ndb.interfaces[ifname].show('json'))
    (ndb.interfaces[ifname].remove().commit())



from pyroute2.netlink import (
    NLA_F_NESTED,
    NLM_F_ACK,
    NLM_F_REQUEST,
    genlmsg,
    nla,
)
from pyroute2.netlink.exceptions import NetlinkError
from pyroute2.netlink.generic import (
    AsyncGenericNetlinkSocket,
    GenericNetlinkSocket,
)

ETHTOOL_GENL_NAME = "ethtool"
ETHTOOL_GENL_VERSION = 1

ETHTOOL_MSG_USER_NONE = 0
ETHTOOL_MSG_STRSET_GET = 1
ETHTOOL_MSG_LINKINFO_GET = 2
ETHTOOL_MSG_LINKINFO_SET = 3
ETHTOOL_MSG_LINKMODES_GET = 4
ETHTOOL_MSG_LINKMODES_SET = 5
ETHTOOL_MSG_LINKSTATE_GET = 6
ETHTOOL_MSG_DEBUG_GET = 7
ETHTOOL_MSG_DEBUG_SET = 8
ETHTOOL_MSG_WOL_GET = 9
ETHTOOL_MSG_WOL_SET = 10
ETHTOOL_MSG_FEATURES_GET = 11
ETHTOOL_MSG_FEATURES_SET = 12
ETHTOOL_MSG_PRIVFLAGS_GET = 13
ETHTOOL_MSG_PRIVFLAGS_SET = 14
ETHTOOL_MSG_RINGS_GET = 15
ETHTOOL_MSG_RINGS_SET = 16


class ethtoolheader(nla):
    nla_flags = NLA_F_NESTED
    nla_map = (
        ('ETHTOOL_A_HEADER_UNSPEC', 'none'),
        ('ETHTOOL_A_HEADER_DEV_INDEX', 'uint32'),
        ('ETHTOOL_A_HEADER_DEV_NAME', 'asciiz'),
        ('ETHTOOL_A_HEADER_FLAGS', 'uint32'),
    )


class ethtoolbitset(nla):
    nla_flags = NLA_F_NESTED
    nla_map = (
        ("ETHTOOL_A_BITSET_UNSPEC", 'none'),
        ("ETHTOOL_A_BITSET_NOMASK", "flag"),
        ("ETHTOOL_A_BITSET_SIZE", 'uint32'),
        ("ETHTOOL_A_BITSET_BITS", 'bitset_bits'),
        ("ETHTOOL_A_BITSET_VALUE", 'hex'),
        ("ETHTOOL_A_BITSET_MASK", 'hex'),
    )

    class bitset_bits(nla):
        nla_flags = NLA_F_NESTED
        nla_map = (
            ('ETHTOOL_A_BITSET_BIT_UNSPEC', 'none'),
            ('ETHTOOL_A_BITSET_BITS_BIT', 'bitset_bits_bit'),
        )

        class bitset_bits_bit(nla):
            nla_flags = NLA_F_NESTED
            nla_map = (
                ('ETHTOOL_A_BITSET_BIT_UNSPEC', 'none'),
                ('ETHTOOL_A_BITSET_BIT_INDEX', 'uint32'),
                ('ETHTOOL_A_BITSET_BIT_NAME', 'asciiz'),
                ('ETHTOOL_A_BITSET_BIT_VALUE', 'flag'),
            )


class ethtool_strset_msg(genlmsg):
    nla_map = (
        ('ETHTOOL_A_STRSET_UNSPEC', 'none'),
        ('ETHTOOL_A_STRSET_HEADER', 'ethtoolheader'),
        ('ETHTOOL_A_STRSET_STRINGSETS', 'strings_sets'),
        ('ETHTOOL_A_STRSET_COUNTS_ONLY', 'flag'),
    )

    ethtoolheader = ethtoolheader

    class strings_sets(nla):
        nla_flags = NLA_F_NESTED
        nla_map = (
            ('ETHTOOL_A_STRINGSETS_UNSPEC', 'none'),
            ('ETHTOOL_A_STRINGSETS_STRINGSET', 'string_set'),
        )

        class string_set(nla):
            nla_flags = NLA_F_NESTED
            nla_map = (
                ('ETHTOOL_A_STRINGSET_UNSPEC', 'none'),
                ('ETHTOOL_A_STRINGSET_ID', 'uint32'),
                ('ETHTOOL_A_STRINGSET_COUNT', 'uint32'),
                ('ETHTOOL_A_STRINGSET_STRINGS', 'stringset_strings'),
            )

            class stringset_strings(nla):
                nla_flags = NLA_F_NESTED
                nla_map = (
                    ('ETHTOOL_A_STRINGS_UNSPEC', 'none'),
                    ('ETHTOOL_A_STRINGS_STRING', 'strings_string'),
                )

                class strings_string(nla):
                    nla_flags = NLA_F_NESTED
                    nla_map = (
                        ('ETHTOOL_A_STRING_UNSPEC', 'none'),
                        ('ETHTOOL_A_STRING_INDEX', 'uint32'),
                        ('ETHTOOL_A_STRING_VALUE', 'asciiz'),
                    )


class ethtool_linkinfo_msg(genlmsg):
    nla_map = (
        ('ETHTOOL_A_LINKINFO_UNSPEC', 'none'),
        ('ETHTOOL_A_LINKINFO_HEADER', 'ethtoolheader'),
        ('ETHTOOL_A_LINKINFO_PORT', 'uint8'),
        ('ETHTOOL_A_LINKINFO_PHYADDR', 'uint8'),
        ('ETHTOOL_A_LINKINFO_TP_MDIX', 'uint8'),
        ('ETHTOOL_A_LINKINFO_TP_MDIX_CTR', 'uint8'),
        ('ETHTOOL_A_LINKINFO_TRANSCEIVER', 'uint8'),
    )

    ethtoolheader = ethtoolheader


class ethtool_linkmode_msg(genlmsg):
    nla_map = (
        ('ETHTOOL_A_LINKMODES_UNSPEC', 'none'),
        ('ETHTOOL_A_LINKMODES_HEADER', 'ethtoolheader'),
        ('ETHTOOL_A_LINKMODES_AUTONEG', 'uint8'),
        ('ETHTOOL_A_LINKMODES_OURS', 'ethtoolbitset'),
        ('ETHTOOL_A_LINKMODES_PEER', 'ethtoolbitset'),
        ('ETHTOOL_A_LINKMODES_SPEED', 'uint32'),
        ('ETHTOOL_A_LINKMODES_DUPLEX', 'uint8'),
    )

    ethtoolheader = ethtoolheader
    ethtoolbitset = ethtoolbitset


class ethtool_linkstate_msg(genlmsg):
    nla_map = (
        ('ETHTOOL_A_LINKSTATE_UNSPEC', 'none'),
        ('ETHTOOL_A_LINKSTATE_HEADER', 'ethtoolheader'),
        ('ETHTOOL_A_LINKSTATE_LINK', 'uint8'),
    )

    ethtoolheader = ethtoolheader


class ethtool_wol_msg(genlmsg):
    nla_map = (
        ('ETHTOOL_A_WOL_UNSPE', 'none'),
        ('ETHTOOL_A_WOL_HEADER', 'ethtoolheader'),
        ('ETHTOOL_A_WOL_MODES', 'ethtoolbitset'),
        ('ETHTOOL_A_WOL_SOPASS', 'hex'),
    )

    ethtoolheader = ethtoolheader
    ethtoolbitset = ethtoolbitset


class ethtool_rings_msg(genlmsg):
    nla_map = (
        ('ETHTOOL_A_RINGS_UNSPEC', 'none'),
        ('ETHTOOL_A_RINGS_HEADER', 'ethtoolheader'),
        ('ETHTOOL_A_RINGS_RX_MAX', 'uint32'),
        ('ETHTOOL_A_RINGS_RX_MINI_MAX', 'uint32'),
        ('ETHTOOL_A_RINGS_RX_JUMBO_MAX', 'uint32'),
        ('ETHTOOL_A_RINGS_TX_MAX', 'uint32'),
        ('ETHTOOL_A_RINGS_RX', 'uint32'),
        ('ETHTOOL_A_RINGS_RX_MINI', 'uint32'),
        ('ETHTOOL_A_RINGS_RX_JUMBO', 'uint32'),
        ('ETHTOOL_A_RINGS_TX', 'uint32'),
        ('ETHTOOL_A_RINGS_RX_BUF_LEN', 'uint32'),
        ('ETHTOOL_A_RINGS_TCP_DATA_SPLIT', 'uint8'),
        ('ETHTOOL_A_RINGS_CQE_SIZE', 'uint32'),
        ('ETHTOOL_A_RINGS_TX_PUSH', 'uint8'),
        ('ETHTOOL_A_RINGS_RX_PUSH', 'uint8'),
        ('ETHTOOL_A_RINGS_TX_PUSH_BUF_LEN', 'uint32'),
        ('ETHTOOL_A_RINGS_TX_PUSH_BUF_LEN_MAX', 'uint32'),
    )

    ethtoolheader = ethtoolheader


class AsyncNlEthtool(AsyncGenericNetlinkSocket):
    async def _do_request(self, msg, msg_flags=NLM_F_REQUEST):
        return await self.nlm_request(
            msg, msg_type=self.prid, msg_flags=msg_flags
        )

    async def is_nlethtool_in_kernel(self):
        try:
            await self.bind(ETHTOOL_GENL_NAME, ethtool_linkinfo_msg)
        except NetlinkError:
            return False
        return True

    def _get_dev_header(self, ifname=None, ifindex=None):
        if ifindex is not None:
            return {'attrs': [['ETHTOOL_A_HEADER_DEV_INDEX', ifindex]]}
        elif ifname is not None:
            return {'attrs': [['ETHTOOL_A_HEADER_DEV_NAME', ifname]]}
        else:
            raise ValueError("Need ifname or ifindex")

    async def get_linkinfo(self, ifname=None, ifindex=None):
        msg = ethtool_linkinfo_msg()
        msg["cmd"] = ETHTOOL_MSG_LINKINFO_GET
        msg['version'] = ETHTOOL_GENL_VERSION
        msg["attrs"].append(
            (
                'ETHTOOL_A_LINKINFO_HEADER',
                self._get_dev_header(ifname, ifindex),
            )
        )

        await self.bind(ETHTOOL_GENL_NAME, ethtool_linkinfo_msg)
        return await self._do_request(msg)

    async def get_linkmode(self, ifname=None, ifindex=None):
        msg = ethtool_linkmode_msg()
        msg["cmd"] = ETHTOOL_MSG_LINKMODES_GET
        msg['version'] = ETHTOOL_GENL_VERSION
        msg["attrs"].append(
            (
                'ETHTOOL_A_LINKMODES_HEADER',
                self._get_dev_header(ifname, ifindex),
            )
        )

        await self.bind(ETHTOOL_GENL_NAME, ethtool_linkmode_msg)
        return await self._do_request(msg)

    async def get_stringset(self, ifname=None, ifindex=None):
        msg = ethtool_strset_msg()
        msg["cmd"] = ETHTOOL_MSG_STRSET_GET
        msg['version'] = ETHTOOL_GENL_VERSION
        msg["attrs"].append(
            ('ETHTOOL_A_STRSET_HEADER', self._get_dev_header(ifname, ifindex))
        )

        await self.bind(ETHTOOL_GENL_NAME, ethtool_strset_msg)
        return await self._do_request(msg)

    async def get_linkstate(self, ifname=None, ifindex=None):
        msg = ethtool_linkstate_msg()
        msg["cmd"] = ETHTOOL_MSG_LINKSTATE_GET
        msg['version'] = ETHTOOL_GENL_VERSION
        msg["attrs"].append(
            (
                'ETHTOOL_A_LINKSTATE_HEADER',
                self._get_dev_header(ifname, ifindex),
            )
        )

        await self.bind(ETHTOOL_GENL_NAME, ethtool_linkstate_msg)
        return await self._do_request(msg)

    async def get_wol(self, ifname=None, ifindex=None):
        msg = ethtool_wol_msg()
        msg["cmd"] = ETHTOOL_MSG_WOL_GET
        msg['version'] = ETHTOOL_GENL_VERSION
        msg["attrs"].append(
            ('ETHTOOL_A_WOL_HEADER', self._get_dev_header(ifname, ifindex))
        )

        await self.bind(ETHTOOL_GENL_NAME, ethtool_wol_msg)
        return await self._do_request(msg)

    async def get_rings(self, ifname=None, ifindex=None):
        msg = ethtool_rings_msg()
        msg["cmd"] = ETHTOOL_MSG_RINGS_GET
        msg["version"] = ETHTOOL_GENL_VERSION
        msg["attrs"].append(
            ('ETHTOOL_A_RINGS_HEADER', self._get_dev_header(ifname, ifindex))
        )

        await self.bind(ETHTOOL_GENL_NAME, ethtool_rings_msg)
        return await self._do_request(msg)

    async def set_rings(self, rings, ifname=None, ifindex=None):
        rings["cmd"] = ETHTOOL_MSG_RINGS_SET
        rings["version"] = ETHTOOL_GENL_VERSION
        rings["attrs"].append(
            ('ETHTOOL_A_RINGS_HEADER', self._get_dev_header(ifname, ifindex))
        )

        await self.bind(ETHTOOL_GENL_NAME, ethtool_rings_msg)
        return await self._do_request(
            rings, msg_flags=NLM_F_REQUEST | NLM_F_ACK
        )


class NlEthtool(GenericNetlinkSocket):

    async_class = AsyncNlEthtool

    def is_nlethtool_in_kernel(self):
        return self._run_with_cleanup(self.asyncore.is_nlethtool_in_kernel)

    def get_linkinfo(self, ifname=None, ifindex=None):
        return self._run_sync_cleanup(
            self.asyncore.get_linkinfo, ifname, ifindex
        )

    def get_linkmode(self, ifname=None, ifindex=None):
        return self._run_sync_cleanup(
            self.asyncore.get_linkmode, ifname, ifindex
        )

    def get_stringset(self, ifname=None, ifindex=None):
        return self._run_sync_cleanup(
            self.asyncore.get_stringset, ifname, ifindex
        )

    def get_linkstate(self, ifname=None, ifindex=None):
        return self._run_sync_cleanup(
            self.asyncore.get_linkstate, ifname, ifindex
        )

    def get_wol(self, ifname=None, ifindex=None):
        return self._run_sync_cleanup(self.asyncore.get_wol, ifname, ifindex)

    def get_rings(self, ifname=None, ifindex=None):
        return self._run_sync_cleanup(self.asyncore.get_rings, ifname, ifindex)

    def set_rings(self, rings, ifname=None, ifindex=None):
        return self._run_sync_cleanup(
            self.asyncore.set_rings, rings, ifname, ifindex
        )



'''
RTNetlink: network setup
========================

RTNL is a netlink protocol, used to get and set information
about different network objects -- addresses, routes, interfaces
etc.

RTNL protocol-specific data in messages depends on the object
type. E.g., complete packet with the interface address information::

    nlmsg header:
        + uint32 length
        + uint16 type
        + uint16 flags
        + uint32 sequence number
        + uint32 pid
    ifaddrmsg structure:
        + unsigned char ifa_family
        + unsigned char ifa_prefixlen
        + unsigned char ifa_flags
        + unsigned char ifa_scope
        + uint32 ifa_index
    [ optional NLA tree ]

NLA for this kind of packets can be of type IFA_ADDRESS, IFA_LOCAL
etc. -- please refer to the corresponding source.

Other objects types require different structures, sometimes really
complex. All these structures are described in sources.

---------------------------

Module contents:

'''

from pyroute2.common import map_namespace

# Useful RTNL related constants from the kernel sources:
IFNAMSIZ = 16
ALTIFNAMSIZ = 128

#  RTnetlink multicast group flags (for use with bind())
RTMGRP_NONE = 0x0
RTMGRP_LINK = 0x1
RTMGRP_NOTIFY = 0x2
RTMGRP_NEIGH = 0x4
RTMGRP_TC = 0x8
RTMGRP_IPV4_IFADDR = 0x10
RTMGRP_IPV4_MROUTE = 0x20
RTMGRP_IPV4_ROUTE = 0x40
RTMGRP_IPV4_RULE = 0x80
RTMGRP_IPV6_IFADDR = 0x100
RTMGRP_IPV6_MROUTE = 0x200
RTMGRP_IPV6_ROUTE = 0x400
RTMGRP_IPV6_IFINFO = 0x800
RTMGRP_DECnet_IFADDR = 0x1000
RTMGRP_NOP2 = 0x2000
RTMGRP_DECnet_ROUTE = 0x4000
RTMGRP_DECnet_RULE = 0x8000
RTMGRP_NOP4 = 0x10000
RTMGRP_IPV6_PREFIX = 0x20000
RTMGRP_IPV6_RULE = 0x40000
RTMGRP_MPLS_ROUTE = 0x4000000

# multicast group ids (for use with {add,drop}_membership)
RTNLGRP_NONE = 0
RTNLGRP_LINK = 1
RTNLGRP_NOTIFY = 2
RTNLGRP_NEIGH = 3
RTNLGRP_TC = 4
RTNLGRP_IPV4_IFADDR = 5
RTNLGRP_IPV4_MROUTE = 6
RTNLGRP_IPV4_ROUTE = 7
RTNLGRP_IPV4_RULE = 8
RTNLGRP_IPV6_IFADDR = 9
RTNLGRP_IPV6_MROUTE = 10
RTNLGRP_IPV6_ROUTE = 11
RTNLGRP_IPV6_IFINFO = 12
RTNLGRP_DECnet_IFADDR = 13
RTNLGRP_NOP2 = 14
RTNLGRP_DECnet_ROUTE = 15
RTNLGRP_DECnet_RULE = 16
RTNLGRP_NOP4 = 17
RTNLGRP_IPV6_PREFIX = 18
RTNLGRP_IPV6_RULE = 19
RTNLGRP_ND_USEROPT = 20
RTNLGRP_PHONET_IFADDR = 21
RTNLGRP_PHONET_ROUTE = 22
RTNLGRP_DCB = 23
RTNLGRP_IPV4_NETCONF = 24
RTNLGRP_IPV6_NETCONF = 25
RTNLGRP_MDB = 26
RTNLGRP_MPLS_ROUTE = 27
RTNLGRP_NSID = 28
RTNLGRP_MPLS_NETCONF = 29
RTNLGRP_IPV4_MROUTE_R = 30
RTNLGRP_IPV6_MROUTE_R = 31

# Types of messages
# RTM_BASE = 16
RTM_NEWLINK = 16
RTM_DELLINK = 17
RTM_GETLINK = 18
RTM_SETLINK = 19
RTM_NEWADDR = 20
RTM_DELADDR = 21
RTM_GETADDR = 22
RTM_NEWROUTE = 24
RTM_DELROUTE = 25
RTM_GETROUTE = 26
RTM_NEWNEIGH = 28
RTM_DELNEIGH = 29
RTM_GETNEIGH = 30
RTM_NEWRULE = 32
RTM_DELRULE = 33
RTM_GETRULE = 34
RTM_NEWQDISC = 36
RTM_DELQDISC = 37
RTM_GETQDISC = 38
RTM_NEWTCLASS = 40
RTM_DELTCLASS = 41
RTM_GETTCLASS = 42
RTM_NEWTFILTER = 44
RTM_DELTFILTER = 45
RTM_GETTFILTER = 46
RTM_NEWACTION = 48
RTM_DELACTION = 49
RTM_GETACTION = 50
RTM_NEWPREFIX = 52
RTM_GETMULTICAST = 58
RTM_GETANYCAST = 62
RTM_NEWNEIGHTBL = 64
RTM_GETNEIGHTBL = 66
RTM_SETNEIGHTBL = 67
RTM_NEWNDUSEROPT = 68
RTM_NEWADDRLABEL = 72
RTM_DELADDRLABEL = 73
RTM_GETADDRLABEL = 74
RTM_GETDCB = 78
RTM_SETDCB = 79
RTM_NEWNETCONF = 80
RTM_DELNETCONF = 81
RTM_GETNETCONF = 82
RTM_NEWMDB = 84
RTM_DELMDB = 85
RTM_GETMDB = 86
RTM_NEWNSID = 88
RTM_DELNSID = 89
RTM_GETNSID = 90
RTM_NEWSTATS = 92
RTM_GETSTATS = 94
RTM_NEWCACHEREPORT = 96
RTM_NEWLINKPROP = 108
RTM_DELLINKPROP = 109
RTM_GETLINKPROP = 110
# fake internal message types
RTM_NEWNETNS = 500
RTM_DELNETNS = 501
RTM_GETNETNS = 502
RTM_NEWPROBE = 504
RTM_DELPROBE = 505
RTM_GETPROBE = 506
(RTM_NAMES, RTM_VALUES) = map_namespace('RTM_', globals())

TC_H_INGRESS = 0xFFFFFFF1
TC_H_CLSACT = TC_H_INGRESS
TC_H_ROOT = 0xFFFFFFFF


RTMGRP_DEFAULTS = (
    RTMGRP_IPV4_IFADDR
    | RTMGRP_IPV6_IFADDR
    | RTMGRP_IPV4_ROUTE
    | RTMGRP_IPV6_ROUTE
    | RTMGRP_IPV4_RULE
    | RTMGRP_IPV6_RULE
    | RTMGRP_NEIGH
    | RTMGRP_LINK
    | RTMGRP_TC
    | RTMGRP_MPLS_ROUTE
)

encap_type = {'unspec': 0, 'mpls': 1, 0: 'unspec', 1: 'mpls'}

rtypes = {
    'RTN_UNSPEC': 0,
    'RTN_UNICAST': 1,  # Gateway or direct route
    'RTN_LOCAL': 2,  # Accept locally
    'RTN_BROADCAST': 3,  # Accept locally as broadcast
    #                        send as broadcast
    'RTN_ANYCAST': 4,  # Accept locally as broadcast,
    #                        but send as unicast
    'RTN_MULTICAST': 5,  # Multicast route
    'RTN_BLACKHOLE': 6,  # Drop
    'RTN_UNREACHABLE': 7,  # Destination is unreachable
    'RTN_PROHIBIT': 8,  # Administratively prohibited
    'RTN_THROW': 9,  # Not in this table
    'RTN_NAT': 10,  # Translate this address
    'RTN_XRESOLVE': 11,
}  # Use external resolver
# normalized
rt_type = dict(
    [(x[0][4:].lower(), x[1]) for x in rtypes.items()]
    + [(x[1], x[0][4:].lower()) for x in rtypes.items()]
)

rtprotos = {
    'RTPROT_UNSPEC': 0,
    'RTPROT_REDIRECT': 1,  # Route installed by ICMP redirects;
    #                        not used by current IPv4
    'RTPROT_KERNEL': 2,  # Route installed by kernel
    'RTPROT_BOOT': 3,  # Route installed during boot
    'RTPROT_STATIC': 4,  # Route installed by administrator
    # Values of protocol >= RTPROT_STATIC are not
    # interpreted by kernel;
    # keep in sync with iproute2 !
    'RTPROT_GATED': 8,  # gated
    'RTPROT_RA': 9,  # RDISC/ND router advertisements
    'RTPROT_MRT': 10,  # Merit MRT
    'RTPROT_ZEBRA': 11,  # Zebra
    'RTPROT_BIRD': 12,  # BIRD
    'RTPROT_DNROUTED': 13,  # DECnet routing daemon
    'RTPROT_XORP': 14,  # XORP
    'RTPROT_NTK': 15,  # Netsukuku
    'RTPROT_DHCP': 16,
}  # DHCP client
# normalized
rt_proto = dict(
    [(x[0][7:].lower(), x[1]) for x in rtprotos.items()]
    + [(x[1], x[0][7:].lower()) for x in rtprotos.items()]
)

rtscopes = {
    'RT_SCOPE_UNIVERSE': 0,
    'RT_SCOPE_SITE': 200,
    'RT_SCOPE_LINK': 253,
    'RT_SCOPE_HOST': 254,
    'RT_SCOPE_NOWHERE': 255,
}
# normalized
rt_scope = dict(
    [(x[0][9:].lower(), x[1]) for x in rtscopes.items()]
    + [(x[1], x[0][9:].lower()) for x in rtscopes.items()]
)



from secrets import SystemRandom
from typing import Any, Optional, Union

from pyroute2.dhcp.fsm import State

random = SystemRandom()


def random_xid_prefix() -> int:
    '''A (max) 32 bit random int with its lowest nibble set to 0.

    These last 4 bits are used by the client to store its state, in the form
    of the associated `State` value, when sending a DHCP message.
    '''
    return random.randint(0x00000010, 0xFFFFFFF0)


class Xid:
    '''Transaction IDs used to identify responses to DHCP requests.

    We use the last nibble to store the state the message was sent in.

    The most significant bits store a random value we use to match requests
    and responses, since the RFC specifies that the server must send back
    the same value when answering. (see RFC section 4.1)
    '''

    def __init__(self, value: Optional[Union[int, 'Xid']] = None):
        if isinstance(value, int):
            int_value = value
        elif isinstance(value, Xid):
            int_value = int(value)
        elif value is None:
            int_value = random_xid_prefix()
        else:
            raise TypeError(f'{value!r} is not an xid')
        assert int_value < 0xFFFFFFFF  # we have 32 bits
        self._value = int_value

    @property
    def random_part(self) -> int:
        '''The random part of the transaction id.'''
        return self._value & 0xFFFFFFF0

    @property
    def request_state(self) -> Optional[State]:
        '''The state in which the request was sent.

        Since servers answer with the same transaction ID as the request,
        we can use this to know what client state does a response answer to.
        '''
        try:
            return State(self._value & 0xF)
        except ValueError:
            return None

    def for_state(self, state: State) -> 'Xid':
        '''A new Xid built from the random part + the state.'''
        return Xid(self.random_part | state)

    def __index__(self) -> int:
        '''Allows xids to be used as int.'''
        return self._value

    def matches(self, received_xid: 'Xid'):
        '''Loose match, whether the random part of both XIDs match.

        This can be used to check if a message is indeed in response
        to a request we sent.
        '''
        return self.random_part == received_xid.random_part

    def __eq__(self, other: Any) -> bool:
        '''Xids compare to other xids or ints.'''
        if isinstance(other, Xid):
            return self._value == other._value
        elif isinstance(other, int):
            return self._value == other
        return False

    def __str__(self) -> str:
        return hex(self._value)

    def __repr__(self) -> str:
        return f"Xid({self})"
